// broken_indicator.cc
//
//   Copyright (C) 2005 Daniel Burrows
//
//   This program is free software; you can redistribute it and/or
//   modify it under the terms of the GNU General Public License as
//   published by the Free Software Foundation; either version 2 of
//   the License, or (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//   General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; see the file COPYING.  If not, write to
//   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA 02111-1307, USA.

#include "broken_indicator.h"

#include <aptitude.h>

#include <generic/apt.h>
#include <generic/aptcache.h>
#include <generic/aptitude_resolver.h>

#include <vscreen/config/colors.h>
#include <vscreen/config/keybindings.h>
#include <vscreen/fragment.h>
#include <vscreen/vs_text_layout.h>

#include <apt-pkg/pkgsystem.h>

#include <string>
#include <vector>

using namespace std;

/** A simple indicator, usually placed at the bottom of the screen,
 *  that describes the current state of the problem resolver.  Hidden
 *  if no problem resolver is active.
 *
 *  \todo write a simple table fragment class and use that to
 *  right-justify the text that obviously should be
 */
class broken_indicator:public vs_text_layout
{
  aptitude_resolver::solution last_sol;

  void handle_cache_reload()
  {
    if(apt_cache_file)
      {
	(*apt_cache_file)->package_state_changed.connect(sigc::mem_fun(*this, &broken_indicator::update));
	(*apt_cache_file)->selected_solution_changed.connect(sigc::mem_fun(*this, &broken_indicator::update));
      }

    update();
  }
public:
  broken_indicator()
  {
    if(apt_cache_file)
      {
	(*apt_cache_file)->package_state_changed.connect(sigc::mem_fun(*this, &broken_indicator::update));
	(*apt_cache_file)->selected_solution_changed.connect(sigc::mem_fun(*this, &broken_indicator::update));
      }

    cache_closed.connect(sigc::mem_fun(*this, &broken_indicator::update));
    cache_reloaded.connect(sigc::mem_fun(*this, &broken_indicator::handle_cache_reload));

    set_bg_style(get_style("Error"));

    update();
  }

  void update()
  {
    if((!apt_cache_file) || !(*apt_cache_file)->resolver_exists())
      {
	set_fragment(fragf(""));
	last_sol.nullify();
	hide();
	return;
      }

    wstring next=global_bindings.readable_keyname("NextSolution");
    wstring prev=global_bindings.readable_keyname("PrevSolution");
    wstring examine=global_bindings.readable_keyname("ExamineSolution");
    wstring apply=global_bindings.readable_keyname("ApplySolution");

    // try to retrieve the current solution to show information about
    // it.
    try
      {
	aptitude_resolver::solution sol=(*apt_cache_file)->get_current_solution();

	if(sol == last_sol)
	  return;

	last_sol=sol;

	if(sol.get_actions().empty())
	  {
	    set_fragment(fragf("%s", _("Internal error: unexpected null solution.")));
	    show();
	    return;
	  }

	int install_count=0, remove_count=0, keep_count=0, upgrade_count=0, downgrade_count=0;

	for(std::map<aptitude_resolver::package,
	      aptitude_resolver::action>::const_iterator i=sol.get_actions().begin();
	    i!=sol.get_actions().end(); ++i)
	  {
	    pkgCache::PkgIterator pkg=i->first.get_pkg();
	    pkgCache::VerIterator curver=pkg.CurrentVer();
	    pkgCache::VerIterator instver=(*apt_cache_file)[pkg].InstVerIter(*apt_cache_file);
	    pkgCache::VerIterator newver=i->second.ver.get_ver();

	    // If not, we have a problem.
	    assert(instver!=newver);

	    if(newver == curver)
	      ++keep_count;
	    else if(curver.end())
	      ++install_count;
	    else if(newver.end())
	      ++remove_count;
	    else
	      {
		int cmp=_system->VS->CmpVersion(curver.VerStr(),
						newver.VerStr());

		// The versions shouldn't be equal -- otherwise
		// something is majorly wrong.
		// assert(cmp!=0);
		//
		// The above is not true: consider, eg, the case of a
		// locally compiled package and a standard package.

		/** \todo indicate "sidegrades" separately? */
		if(cmp<=0)
		  ++upgrade_count;
		else if(cmp>0)
		  ++downgrade_count;
	      }

	    vector<fragment *> fragments;
	    char buf[512];

	    snprintf(buf, 512, (*apt_cache_file)->solution_generation_complete()?"[%d/%d]":"[%d/...]",
		     (*apt_cache_file)->get_selected_solution()+1,
		     (*apt_cache_file)->generated_solution_count());
	    fragments.push_back(fragf("%s ", buf));

	    fragments.push_back(fragf(_("Suggest ")));

	    vector<fragment *> suggestions;

	    if(install_count>1)
	      {
		snprintf(buf, 512, _("%d installs"),
			 install_count);
		suggestions.push_back(text_fragment(buf));
	      }
	    else if(install_count == 1)
	      suggestions.push_back(text_fragment(_("1 install")));

	    if(remove_count>1)
	      {
		snprintf(buf, 512, _("%d removals"),
			 remove_count);
		suggestions.push_back(text_fragment(buf));
	      }
	    else if(remove_count == 1)
	      suggestions.push_back(text_fragment(_("1 removal")));

	    if(keep_count>1)
	      {
		snprintf(buf, 512, _("%d keeps"),
			 keep_count);
		suggestions.push_back(text_fragment(buf));
	      }
	    else if(keep_count == 1)
	      suggestions.push_back(text_fragment(_("1 keep")));

	    if(upgrade_count>1)
	      {
		snprintf(buf, 512, _("%d upgrades"),
			 upgrade_count);
		suggestions.push_back(text_fragment(buf));
	      }
	    else if(upgrade_count == 1)
	      suggestions.push_back(text_fragment(_("1 upgrade")));

	    if(downgrade_count>1)
	      {
		snprintf(buf, 512, _("%d downgrades"),
			 downgrade_count);
		suggestions.push_back(text_fragment(buf));
	      }
	    else if(downgrade_count == 1)
	      suggestions.push_back(text_fragment(_("1 downgrade")));

	    fragments.push_back(join_fragments(suggestions,
					       L","));

	    fragments.push_back(newline_fragment());

	    vector<fragment *> key_hints;

	    key_hints.push_back(fragf(_("%s: Examine"),
				      examine.c_str()));
	    key_hints.push_back(fragf(_("%s: Apply"),
				      apply.c_str()));
	    key_hints.push_back(fragf(_("%s: Next"),
				      next.c_str()));
	    key_hints.push_back(fragf(_("%s: Previous"),
				      prev.c_str()));

	    fragments.push_back(join_fragments(key_hints,
					       L"  "));

	    fragment *f=sequence_fragment(fragments);
	    set_fragment(hardwrapbox(f));
	  }
      }
    catch(NoMoreTime)
      {
	set_fragment(fragf(_("Solution search failed.  %s: try harder"),
			   next.c_str()));
      }
    catch(NoMoreSolutions)
      {
	set_fragment(fragf(_("Unable to resolve dependencies.")));
      }

    show();
  }
};

vscreen_widget *make_broken_indicator()
{
  return new broken_indicator;
}
