// cmdline_resolver.cc
//
//   Copyright (C) 2005 Daniel Burrows
//
//   This program is free software; you can redistribute it and/or
//   modify it under the terms of the GNU General Public License as
//   published by the Free Software Foundation; either version 2 of
//   the License, or (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//   General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; see the file COPYING.  If not, write to
//   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA 02111-1307, USA.

#include "cmdline_resolver.h"

#include "cmdline_action.h"
#include "cmdline_common.h"
#include "cmdline_prompt.h"
#include "cmdline_show.h"
#include "cmdline_show_broken.h"

#include <aptitude.h>
#include <solution_fragment.h>

#include <generic/aptcache.h>
#include <generic/aptitude_resolver.h>
#include <generic/config_signal.h>

#include <vscreen/fragment.h>

#include <apt-pkg/error.h>

#include <iostream>
#include <fstream>

using namespace std;

static void setup_resolver(pkgset &to_install,
			   pkgset &to_hold,
			   pkgset &to_remove,
			   pkgset &to_purge,
			   bool force_no_change)
{
  (*apt_cache_file)->set_resolver_debug(aptcfg->FindB(PACKAGE "::CmdLine::Resolver-Debug", false));

  // For all packages that the user listed on the command-line (i.e.,
  // all in to_install, to_hold, to_remove, and to_purge), tell the
  // resolver to try *really really hard* to avoid altering their
  // state.
  if(force_no_change && (*apt_cache_file)->resolver_exists())
    {
      pkgset *sets[4]={&to_install, &to_hold, &to_remove, &to_purge};
      int tweak_amt=aptcfg->FindI(PACKAGE "::CmdLine::Request-Strictness", 1000);

      for(int i=0; i<4; ++i)
	{
	  pkgset &S=*sets[i];

	  for(pkgset::const_iterator p=S.begin();
	      p!=S.end(); ++p)
	    {
	      pkgDepCache::StateCache &state=(*apt_cache_file)[*p];
	      pkgCache::VerIterator instver=state.InstVerIter(*apt_cache_file);

	      for(pkgCache::VerIterator v=p->VersionList();
		  !v.end(); ++v)
		if(instver == v)
		  (*apt_cache_file)->tweak_score(*p, v,
						 tweak_amt);

	      if(instver.end())
		(*apt_cache_file)->tweak_score(*p, pkgCache::VerIterator(*apt_cache_file),
					       tweak_amt);
	    }
	}
    }
}

static inline fragment *flowindentbox(int i1, int irest, fragment *f)
{
  return indentbox(i1, irest, flowbox(f));
}

static void resolver_help(ostream &out)
{
  fragment *f=indentbox(2, 2,
			fragf(_("y: %F"
				"n: %F"
				"q: %F"
				",: %F"
				".: %F"
				"<ACTION> pkg... : %F%n"
				"%F"
				"%F"
				"%F"
				"%F"
				"%F"
				"%F"
				"%F"
				"%F"),
			      flowindentbox(0, 3,
					    fragf(_("accept the proposed changes"))),
			      flowindentbox(0, 3,
					    fragf(_("reject the proposed changes and search for another solution"))),
			      flowindentbox(0, 3,
					    fragf(_("give up and quit the program"))),
			      flowindentbox(0, 3,
					    fragf(_("move to the next solution"))),
			      flowindentbox(0, 3,
					    fragf(_("move to the previous solution"))),
			      flowindentbox(0, 3,
					    fragf(_("adjust the state of the listed packages, where ACTION is one of:"))),
			      flowindentbox(0, 4,
					    fragf(_("'+' to install packages"))),
			      flowindentbox(0, 5,
					    fragf(_("'+M' to install packages and immediately flag them as automatically installed"))),
			      flowindentbox(0, 4,
					    fragf(_("'-' to remove packages"))),
			      flowindentbox(0, 4,
					    fragf(_("'_' to purge packages"))),
			      flowindentbox(0, 4,
					    fragf(_("'=' to place packages on hold"))),
			      flowindentbox(0, 4,
					    fragf(_("':' to keep packages in their current state without placing them on hold"))),
			      flowindentbox(0, 4,
					    fragf(_("'&M' to mark packages as automatically installed"))),
			      flowindentbox(0, 4,
					    fragf(_("'&m' to mark packages as manually installed"))),
			      flowindentbox(0, 3,
					    fragf(_("Adjustments will cause the current solution to be discarded and recalculated as necessary.")))));

  out << f->layout(screen_width, screen_width, style());
  delete f;
}

bool cmdline_resolve_deps(pkgset &to_install,
			  pkgset &to_hold,
			  pkgset &to_remove,
			  pkgset &to_purge,
			  bool assume_yes,
			  bool force_no_change,
			  int verbose)
{
  while(!show_broken())
    {
      setup_resolver(to_install, to_hold, to_remove, to_purge,
		     force_no_change);
      aptitude_resolver::solution lastsol;

      // The inner loop tries to generate solutions until some
      // packages are modified (then the new set of broken packages,
      // if any, is displayed and we start over)
      bool modified_pkgs=false;
      while(!modified_pkgs)
	try
	  {
	    try
	      {
		aptitude_resolver::solution sol=(*apt_cache_file)->get_current_solution();

		if(sol != lastsol)
		  {
		    fragment *f=sequence_fragment(flowbox(text_fragment(_("The following actions will resolve these dependencies:"))),
						  newline_fragment(),
						  solution_fragment(sol),
						  NULL);

		    update_screen_width();

		    fragment_contents lines=f->layout(screen_width, screen_width, style());

		    delete f;

		    cout << lines << endl;
		    lastsol=sol;
		  }

		string response=assume_yes?"Y":prompt_string(_("Accept this solution? [Y/n/q/?] "));

		string::size_type loc=0;
		while(loc<response.size() && isspace(response[loc]))
		  ++loc;
		if(loc == response.size())
		  {
		    response='Y';
		    loc=0;
		  }

		switch(toupper(response[loc]))
		  {
		  case 'Y':
		    (*apt_cache_file)->apply_current_solution(NULL);
		    modified_pkgs=true;
		    break;
		  case 'N':
		    {
		      int curr_count=(*apt_cache_file)->generated_solution_count();

		      // Move to the last cached solution.
		      if(curr_count>0)
			while((*apt_cache_file)->get_selected_solution()<curr_count)
			  (*apt_cache_file)->next_solution();
		    }
		    break;
		  case 'Q':
		    cout << _("Abandoning all efforts to resolve these dependencies.") << endl;
		    return false;
		  case '.':
		    (*apt_cache_file)->next_solution();
		    break;
		  case ',':
		    // As above, errors pop to the outer 'catch'
		    (*apt_cache_file)->previous_solution();
		    break;
		  case '?':
		    cout << _("The following commands are available:") << endl;
		    resolver_help(cout);
		    break;
		  case '+':
		  case '-':
		  case '=':
		  case '_':
		  case ':':
		    cmdline_parse_action(response, to_install, to_hold,
					 to_remove, to_purge, verbose);
		    modified_pkgs=true;
		    break;
		    // Undocumented debug feature:
		  case '~':
		    {
		      string fn=prompt_string(_("File to write resolver state to: "));
		      ofstream f(fn.c_str());
		      if(!f)
			_error->Errno("dump_resolver", _("Unable to open %s for writing"), fn.c_str());
		      else
			{
			  (*apt_cache_file)->dump_resolver(f);
			  if(!f)
			    _error->Errno("dump_resolver", _("Error writing resolver state to %s"), fn.c_str());
			}
		    }
		    break;
		  default:
		    cout << _("Invalid response; please enter one of the following commands:") << endl;
		    resolver_help(cout);
		    break;
		  }
	      }
	    catch(NoMoreTime)
	      {
		bool done=false;
		while(!done)
		  {
		    string response=prompt_string(_("No solution found within the allotted time.  Try harder? [Y/n]"));

		    string::size_type loc=0;
		    while(loc<response.size() && isspace(response[loc]))
		      ++loc;
		    if(loc == response.size())
		      {
			loc=0;
			response='Y';
		      }
		    switch(toupper(response[loc]))
		      {
		      case 'Y':
			try
			  {
			    (*apt_cache_file)->next_solution();
			    done=true;
			  }
			catch(NoMoreTime)
			  {
			    // ignore and continue looping.
			  }
			// NoMoreExceptions flows to the outer catch.
			break;
		      case 'N':
			cout << _("Abandoning all efforts to resolve these dependencies.") << endl;
			return false;
		      default:
			cout << _("Invalid response; please enter 'y' or 'n'.") << endl;
		      }
		  }
	      }
	  }
	catch(NoMoreSolutions)
	  {
	    if((*apt_cache_file)->generated_solution_count()==0)
	      {
		cout << _("Unable to resolve dependencies!  Giving up...") << endl;
		return false;
	      }
	    else
	      {
		cout << _("No more solutions.") << endl;
		// Force it to re-print the last solution.
		lastsol.nullify();
	      }
	  }
    }

  return true;
}
