// aptitude_resolver.cc
//
//   Copyright (C) 2005 Daniel Burrows
//
//   This program is free software; you can redistribute it and/or
//   modify it under the terms of the GNU General Public License as
//   published by the Free Software Foundation; either version 2 of
//   the License, or (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//   General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; see the file COPYING.  If not, write to
//   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA 02111-1307, USA.

#include "aptitude_resolver.h"

aptitude_resolver::aptitude_resolver(int step_penalty,
				     int broken_penalty,
				     int infinity,
				     int max_successors,
				     int resolution_score,
				     aptitudeDepCache *cache)
  :generic_problem_resolver<aptitude_universe>(step_penalty, broken_penalty, infinity, max_successors, resolution_score, aptitude_universe(cache))
{
}

void aptitude_resolver::add_action_scores(int preserve_score, int auto_score,
					  int remove_score, int keep_score,
					  int install_score, int upgrade_score,
					  int non_default_score, int essential_remove)
{
  // Should I stick with APT iterators instead?  This is a bit more
  // convenient, though..
  for(aptitude_universe::package_iterator pi=aptitude_universe::package_iterator(get_universe().get_cache()->PkgBegin(), get_universe().get_cache());
      !pi.end(); ++pi)
    {
      const aptitude_universe::package &p=*pi;
      aptitudeDepCache::aptitude_state &state=get_universe().get_cache()->get_ext_state(p.get_pkg());

      for(aptitude_universe::package::version_iterator vi=p.versions_begin(); !vi.end(); ++vi)
	{
	  aptitude_universe::version v=*vi;

	  // Remember, the "current version" is the InstVer.
	  if(v==p.current_version())
	    {
	      bool is_user_choice=false;
	      if(!v.get_ver().end() && state.install_reason == aptitudeDepCache::manual)
		is_user_choice=true;
	      else if(v.get_ver().end() && state.remove_reason == aptitudeDepCache::manual)
		is_user_choice=true;

	      if(is_user_choice)
		add_version_score(v, preserve_score);
	      else
		add_version_score(v, auto_score);
	    }
	  // Ok, if this version is selected it'll be a change.
	  else if(v.get_ver()==p.get_pkg().CurrentVer())
	    add_version_score(v, keep_score);
	  else if(v.get_ver().end())
	    add_version_score(v, remove_score);
	  else if(v.get_ver()==(*get_universe().get_cache())[p.get_pkg()].CandidateVerIter(*get_universe().get_cache()))
	    {
	      // Could try harder not to break holds.
	      if(p.get_pkg().CurrentVer().end())
		add_version_score(v, install_score);
	      else
		add_version_score(v, upgrade_score);
	    }
	  else
	    // We know that:
	    //  - this version wasn't requrested by the user
	    //  - it's not the current version
	    //  - it's not the candidate version
	    //  - it's not a removal
	    //  - it follows that this is a non-default version.
	    add_version_score(v, non_default_score);

	  // In addition, add the essential-remoal score:
	  if((p.get_pkg()->Flags & pkgCache::Flag::Essential) &&
	     v.get_ver().end())
	    add_version_score(v, essential_remove);
	}
    }
}

void aptitude_resolver::add_priority_scores(int important,
					    int required,
					    int standard,
					    int optional,
					    int extra)
{
  for(aptitude_universe::package_iterator pi=aptitude_universe::package_iterator(get_universe().get_cache()->PkgBegin(), get_universe().get_cache());
      !pi.end(); ++pi)
    for(aptitude_universe::package::version_iterator vi=(*pi).versions_begin(); !vi.end(); ++vi)
      {
	if(vi.get_ver().end())
	  continue;

	int score_tweak=0;
	switch(vi.get_ver()->Priority)
	  {
	  case pkgCache::State::Important:
	    score_tweak=important;
	    break;
	  case pkgCache::State::Required:
	    score_tweak=required;
	    break;
	  case pkgCache::State::Standard:
	    score_tweak=standard;
	    break;
	  case pkgCache::State::Optional:
	    score_tweak=optional;
	    break;
	  case pkgCache::State::Extra:
	    score_tweak=extra;
	    break;
	  default:
	    // ??????
	    break;
	  }

	add_version_score(*vi, score_tweak);
      }
}

std::ostream &operator<<(ostream &out, aptitude_resolver_dep d)
{
  out << d.get_source().get_package().get_name()
      << " "
      << d.get_source().get_name()
      << " -> {";

  for(aptitude_resolver_dep::solver_iterator i=d.solvers_begin(); !i.end(); ++i)
    {
      if(i!=d.solvers_begin())
	out << " ";
      out << (*i).get_package().get_name() << " "
	  << (*i).get_name();
    }
  out << "}";

  return out;
}
