// problemresolver.h                  -*-c++-*-
//
//   Copyright (C) 2005 Daniel Burrows
//
//   This program is free software; you can redistribute it and/or
//   modify it under the terms of the GNU General Public License as
//   published by the Free Software Foundation; either version 2 of
//   the License, or (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//   General Public License for more details.  You should have
//   received a copy of the GNU General Public License along with this
//   program; see the file COPYING.  If not, write to the Free
//   Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//   MA 02111-1307, USA.

//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; see the file COPYING.  If not, write to
//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//  Boston, MA 02111-1307, USA.
//
//
// Generic problem resolver (generic because I want to be able to do
// regression-testing for once, if I can figure out how, and Packages
// files make lousy regression-tests).

#ifndef PROBLEMRESOLVER_H
#define PROBLEMRESOLVER_H

#include <assert.h>

#include <map>
#include <queue>
#include <set>
#include <vector>

#include <iostream>

/** A dummy iterator that's always an "end" iterator. */
template<class V>
struct dummy_end_iterator
{
public:
  /** \return \b true */
  static bool end() {return true;}
};

/** Aborts: this iterator is always invalid. */
template<class V>
static inline const V &operator*(const dummy_end_iterator<V>&)
{
  abort();
}

/** Aborts. */
template<class V>
static inline dummy_end_iterator<V> operator++(dummy_end_iterator<V>&)
{
  abort();
}

/** \defgroup problemresolver Aptitude's problem resolver
 *
 *  This is a replacement for the generic apt problem resolver.  It
 *  uses a different algorithm which, while it is less efficient,
 *  allows the program to have more direct control over the results,
 *  allows the user to pick and choose from multiple solutions, and
 *  can (eventually) spit out explanations of its actions in a
 *  user-friendly way.
 *
 *  The problem resolver class is templated on abstract packages.
 *  Normally this will be apt packages, but for the purpose of testing
 *  the algorithm a simpler package system is used.  Each package has
 *  an associated "score" which indicates, essentially, how hard the
 *  problem resolver will try to keep it on the system (or, for
 *  negative scores, to remove it from the system).  The sum of all
 *  installed package's scores is used as a heuristic in a directed
 *  search of the solution space (see generic_problem_resolver's
 *  documentation for the interface and its documentation for the gory
 *  details).
 */


/** Generic errors in the problem resolver. */
class ProblemResolverError {
};

/** An exception indicating that no more solutions are available. */
class NoMoreSolutions:public ProblemResolverError {
};

/** Represents a partial or complete solution to a dependency
 *  problem.  Solutions are transparently refcounted to save on
 *  memory and avoid copies.
 */
template<class PackageUniverse>
class generic_solution
{
public:
  typedef typename PackageUniverse::package package;
  typedef typename PackageUniverse::version version;
  typedef typename PackageUniverse::dep dep;

  /** Represents a single action taken by the resolver: the
   *  installation of a particular version of a package.
   */
  struct action
  {
    version ver;

    action() {}

    action(const version &_ver):ver(_ver) {}

    bool operator<(const action &other) const {return ver<other.ver;}
  };

private:
  class solution_rep
  {
    /** The actions performed by this solution.
     *
     *  Originally the plan was to read this off by tracing to the
     *  root -- but it seems likely that this will be read many more
     *  times than it's created, so I'm swallowing the space/time
     *  hit to build a lookup table at each node in the hope that it
     *  pays off later..
     *
     *  Note that this also means you can't find out chronological
     *  ordering, which might help make sense of everything; that
     *  can be added in later one way or another, though.  (eg,
     *  storing the DAG structure after all?)
     */
    std::map<package, action> actions;

    /** The full list of currently broken dependencies. */
    std::set<dep> broken_deps;

    /** A set of versions that have been "locked out" by this
     *	solution.  Primarily used to optimize the common case of
     *	having to "climb a hill" as removals propagate up the
     *	dependency chain.  (this allows us to easily force the search
     *	in the direction it has to go)
     */
    std::set<version> forbidden_versions;

    /** The score of this solution. */
    int score;

    /** The combined score due to package version installations
     *  and distance from the root -- "score" is calculated by
     *  adding the broken-dependency count to this.
     */
    int action_score;

    /** The reference count of this solution. */
    mutable unsigned int refcount;

  public:
    void incref() const {++refcount;}
    void decref() const {assert(refcount>0); if(--refcount==0) delete this;}

    /** Construct a new solution_rep.  The initial reference count
     *  is \b 1.
     *
     *  \param ver the version to install beyond what is in the parent
     *  \param parent the parent solution
     *  \param _broken_deps the dependencies that are broken in this partial solution
     *  \param _score the score of this solution
     *  \param forbidden_iter an APT-style iterator over the set of
     *         newly forbidden versions
     *  \param _action_score the score not due to broken deps
     */
    template<class forbid_iter, class action_iter>
    solution_rep(const action_iter &abegin,
		 const action_iter &aend,
		 const generic_solution &parent,
		 const std::set<dep> &_broken_deps,
		 const forbid_iter &forbidden_iter,
		 int _score, int _action_score)
      :actions(parent.get_actions()), broken_deps(_broken_deps),
       score(_score), action_score(_action_score),
       refcount(1)
    {
      for(forbid_iter i=forbidden_iter; !i.end(); ++i)
	forbidden_versions.insert(*i);

      for(action_iter a=abegin; a!=aend; ++a)
	{
	  assert(actions.find(a->ver.get_package()) == actions.end());
	  actions[a->ver.get_package()] = *a;
	}
    }

    /** Construct a solution with \b no action, the given set of
     *  broken dependencies, and the given score.  Used to generate
     *  the root node of the search.
     */
    solution_rep(const std::set<dep> &_broken_deps,
		 int _score)
      :broken_deps(_broken_deps), score(_score),
       action_score(0), refcount(1)
    {
    }

    const std::set<dep> &get_broken_deps() const
    {
      return broken_deps;
    }

    std::set<dep> &get_broken_deps()
    {
      return broken_deps;
    }

    const std::map<package, action> &get_actions() const
    {
      return actions;
    }

    const std::set<version> &get_forbidden_versions() const
    {
      return forbidden_versions;
    }

    int get_score() const {return score;}
    int get_action_score() const {return action_score;}

    version version_of(const package &pkg) const
    {
      typename std::map<package, action>::const_iterator found=actions.find(pkg);
      if(found == actions.end())
	return pkg.current_version();
      else
	return found->second.ver;
    }

    /** \return true iff this solution touches the given package. */
    bool package_modified(const package &pkg) const
    {
      return actions.find(pkg) != actions.end();
    }
  }; // End solution representation.

  solution_rep *real_soln;
public:
  generic_solution():real_soln(0) {}

  /** Create a solution.
   *
   * \param [abegin,aend) the new actions to enqueue
   *
   * \param parent the parent of this solution
   *
   * \param forbidden_iter an APT-style iterator (i.e., with an end()
   * method) over the set of newly forbidden versions.
   *
   * \param score the total score of this new solution
   *
   * \param action_score the portion of score due to actions
   */
  template<class forbid_iter, class action_iter>
  generic_solution(const action_iter &abegin,
		   const action_iter &aend,
		   const generic_solution &parent,
		   const std::set<dep> &broken_deps,
		   const forbid_iter &forbidden_iter,
		   int score, int action_score)
    :real_soln(new solution_rep(abegin, aend,
				parent, broken_deps,
				forbidden_iter,
				score, action_score))
  {
  }

  generic_solution(const std::set<dep> &broken_deps,
		   int score)
    :real_soln(new solution_rep(broken_deps, score))
  {
  }

  generic_solution(const generic_solution &other)
    :real_soln(other.real_soln)
  {
    if(real_soln)
      real_soln->incref();
  }

  ~generic_solution()
  {
    if(real_soln)
      real_soln->decref();
  }

  /** Make this an invalid reference. */
  void nullify()
  {
    if(real_soln)
      real_soln->decref();
    real_soln=0;
  }

  generic_solution &operator=(const generic_solution &other)
  {
    if(other.real_soln)
      other.real_soln->incref();
    if(real_soln)
      real_soln->decref();
    real_soln=other.real_soln;

    return *this;
  }

  /** \note solutions are compared by pointer identity, not by
   *  value.
   *
   *  \return \b true if the solutions are the same solution.
   */
  bool operator==(const generic_solution &other) const
  {
    return &real_soln == &other.real_soln;
  }

  /** Without this method, some random function that produces
   *  incorrect results is used.
   *
   *  \return \b true if the solutions are not the same solution.
   */
  bool operator!=(const generic_solution &other) const
  {
    return &real_soln != &other.real_soln;
  }

  operator bool() const
  {
    return real_soln != 0;
  }

  const action &operator*() const
  {
    return real_soln->get_action();
  }

  const action *operator->() const
  {
    return &real_soln->get_action();
  }

  const generic_solution &get_parent() const
  {
    return real_soln->get_parent();
  }

  /** \return an immutable list of all the broken dependencies
   *  in this solution.  The list will live at least as long as
   *  the reference from which it was retrieved.
   */
  const std::set<dep> &get_broken() const
  {
    return real_soln->get_broken_deps();
  }

  const std::map<package, action> &get_actions() const
  {
    return real_soln->get_actions();
  }

  const std::set<version> &get_forbidden_versions() const
  {
    return real_soln->get_forbidden_versions();
  }

  /** \return the score of the scolution */
  int get_score() const
  {
    return real_soln->get_score();
  }

  /** \return the score of the solution's actions. */
  int get_action_score() const
  {
    return real_soln->get_action_score();
  }

  version version_of(const package &pkg) const
  {
    return real_soln->version_of(pkg);
  }

  bool package_modified(const package &pkg) const
  {
    return real_soln->package_modified(pkg);
  }

  void dump(std::ostream &out) const
  {
    out << "<";
    for(typename std::map<package, action>::const_iterator i=get_actions().begin();
	i!=get_actions().end(); ++i)
      {
	if(i!=get_actions().begin())
	  out << ", ";
	out << i->first.get_name() << ":=" << i->second.ver.get_name();
      }
    out << ">;[";

    for(typename std::set<dep>::const_iterator i=get_broken().begin();
	i!=get_broken().end(); ++i)
      {
	if(i!=get_broken().begin())
	  out << ", ";
	out << *i;
      }

    out<< "];";

    if(!get_forbidden_versions().empty())
      {
	out << "!!";
	for(typename std::set<version>::const_iterator i=get_forbidden_versions().begin();
	    i!=get_forbidden_versions().end(); ++i)
	  {
	    if(i!=get_forbidden_versions().begin())
	      out << ", ";
	    out << i->get_package().get_name() << " " << i->get_name();
	  }
	out << "!!;";
      }

    out << get_score();
  }
}; // End solution wrapper

/** An exception indicating that the resolver ran out of time to
 *  find a solution.
 */
class NoMoreTime:public ProblemResolverError {
};

/** A generic package problem resolver.  The class PackageUniverse
 *  contains information on how to access packages, package versions,
 *  dependencies (in the specialized form this needs), etc.  It should
 *  export classes that are suitable for use as immediate values (eg,
 *  thin wrappers around pointers/PkgIterators).
 *
 *  The problem resolver views the world in terms of packages,
 *  versions of packages, and dependencies.  A dependency states that
 *  a version of a particular package requires one or more other
 *  packages/versions.  The resolver works by repeatedly fixing single
 *  dependencies; its work queue is priority-ordered according to the
 *  relative "goodness" of the solutions.  The "goodness" of a
 *  solution is affected by:
 *
 *    - the penalty/bonus applied to each package version, which
 *      indicates (roughly) how hard aptitude will try to fix the
 *      problem.
 *
 *    - a penalty is applied as the length of a solution (the number
 *      of packages to be installed/removed) grows.  The idea is to
 *      prioritize simpler solutions first.
 *
 *    - a penalty is applied for broken (strong) dependencies.
 *      This aims at directing the search towards "less broken"
 *      situations.
 *
 *  PackageUniverse::package is the representation of a package; the
 *  main interesting thing about packages is that they have lists of
 *  versions.  Packages can be compared with == and <, they have
 *  unique numerical IDs (retrieved via p.get_id()), and
 *  PackageUniverse::package::version_iterator lets you iterate from
 *  p.versions_begin() to p.versions_end(); dereferencing the iterator
 *  results in a PackageUniverse::version.
 *
 *  A PackageUniverse::version is the object upon which dependencies
 *  operate.  Given a PackageUniverse::version, you can compare it
 *  with == and <, retrieve its corresponding package, or get a
 *  PU::version::revdep_iterator to iterate from v.revdeps_begin() to
 *  v.revdeps_end().  Dereferencing this iterator gives you a PU::dep.
 *
 *  A PU::dep is a dependency of the form "A -> B | C | ...".  Given a
 *  dep, you can compare it with ==, retrieve A with
 *  d.get_source(), and retrieve a PU::dep::solvers_iterator
 *  with d.solvers_begin()/d.solvers_end().  \todo add a function to
 *  efficiently test whether a dep is satisfied using information
 *  available at the concrete package system level.
 *
 *  A PU::dep::solvers_iterator iterates over the list B,C,... above.
 *  Dereferencing it returns a ver_iterator.
 *
 *  \note revdeps need not iterate over \b all reverse dependencies of
 *  a version; it is sufficient that for versions v1,v2, if a
 *  dependency is satisfied by one and not the other, it appears in
 *  the revdeps list of at least one of the versions.  (this allows
 *  apt's conflicts to be handled efficiently)
 */


template<class PackageUniverse>
class generic_problem_resolver
{
public:
  typedef typename PackageUniverse::package package;
  typedef typename PackageUniverse::version version;
  typedef typename PackageUniverse::dep dep;

  typedef generic_solution<PackageUniverse> solution;

  typedef typename solution::action action;

private:
  /** Compares solutions according to their "goodness". */
  struct solution_goodness_compare
  {
    bool operator()(const solution &s1, const solution &s2) const
    {
      return s1.get_score() < s2.get_score();
    }
  };

  /** Compares solutions according to their contents; used to
   *  manage "closed".
   */
  struct solution_contents_compare
  {
    bool operator()(const solution &s1, const solution &s2) const
    {
      // Variation on lexicographical comparison.
      //
      // S1 < S2 if
      //   - S1(p)=S2(p) for all p<p' and S1(p')<S2(p'), where
      //     p,p' \in \dom(S1) \cup dom(S2)
      typename std::map<package,action>::const_iterator i1,i2;
      const std::map<package,action>
	&a1=s1.get_actions(), &a2=s2.get_actions();

      return a1 < a2;
    }
  };

  /** How much to reward long and/or broken solutions.  Typically
   *  negative to penalize such things, or 0 to ignore them.
   */
  int step_score, broken_score;

  /** Solutions whose score is smaller than this value will be
   *  discarded rather than being enqueued.
   */
  int minimum_score;

  /** The "maximum" number of successors to generate for any given
   *  node.  Note, however, that a full set of successors will always
   *  be generated for each broken dependency, so this is not exact.
   *
   *  The theoretical justification is that in order to cover all
   *  solutions, it's sufficient to generate all successors for a
   *  single broken dependency.  This avoids the problem that when a
   *  large number of packages are broken, or when a single package
   *  version with a large number of reverse dependencies is broken,
   *  you can end up searching way too many successor nodes, even if
   *  all the successors are discarded outright.
   *
   *  Unfortunately, only generating one successor has its own
   *  problem: solutions might be generated out-of-order (i.e., worse
   *  solutions before better).  This variable allows you to seek a
   *  "happy medium" by generating a reasonable number of successors,
   *  but not too many.  If a single package breaks a large number of
   *  other packages, then any attempt to fix those packages should
   *  generate the correct successor (which will then pop to the top
   *  of the queue and likely stay there), while if a large number of
   *  unrelated packages are broken, it doesn't matter which successor
   *  goes first.
   */
  unsigned int max_successors;

  /** How much to reward real solutions -- solutions that fix all
   *  dependencies.  Make this big to immediately pick them up, or
   *  small to ignore "bad" solutions (at the risk of running out of
   *  time if no better solution pops up).
   */
  int full_solution_score;

  /** The universe in which we are solving problems. */
  const PackageUniverse universe;

  /** The internal scores to apply to individual package
   *  versions.
   */
  int *version_scores;

  /** If \b true, we have exhausted the list of solutions. */
  bool finished:1;

  bool debug:1;

  /** The working queue: */
  std::priority_queue<solution, std::vector<solution>, solution_goodness_compare> open;

  /** Stores already-seen solutions: */
  std::set<solution, solution_contents_compare> closed;

  /** Stores generated solutions: */
  std::vector<solution> generated_solutions;

  /** Represents a possible successor to the given solution, created
   *  by installing a single 'new' version.
   */
  class partial_solution
  {
    /** The solution this is based on. */
    solution s;
    /** More stuff to install */
    std::map<package, version> installations;
  public:
    partial_solution (const solution &_s)
      :s(_s)
    {
    }

    void install(const package &p, const version &v)
    {
      assert(installations.find(p) == installations.end());

      installations[p]=v;
    }

    /** \return the currently installed version of the given
     *	package.
     */
    version version_of(const package &pkg) const
    {
      typename std::map<package, version>::const_iterator found=installations.find(pkg);

      if(found!=installations.end())
	return found->second;
      else
	return s.version_of(pkg);
    }

    /** Test whether the given version is installed in this solution.
     */
    bool ver_installed(const version &test_ver) const
    {
      return version_of(test_ver.get_package()) == test_ver;
    }
  };

  /** \return \b true if the given solution is "irrelevant": that is,
   *  either it was already generated and placed in the closed queue,
   *  or it includes an already-generated solution as a proper subset.
   */
  bool irrelevant(const solution &s)
  {
    if(closed.find(s) != closed.end())
      return true;

    // The efficiency of this step hinges on the *assumption* that the
    // number of solutions that will be generated is small relative to
    // the number of potential solutions.
    for(typename std::vector<solution>::const_iterator i=generated_solutions.begin();
	i!=generated_solutions.end(); ++i)
      if(includes(s.get_actions().begin(), s.get_actions().end(),
		  i->get_actions().begin(), i->get_actions().end()))
	return true;

    if(s.get_score() < minimum_score)
      {
	if(debug)
	  std::cout << "Not generating solution (infinite badness " << s.get_score() << "<" << minimum_score << ")" << std::endl;
	return true;
      }

    return false;
  }

  /** Generates a successor to s by performing the actions
   *  [abegin,aend).
   */
  template<class forbid_iter, class a_iter>
  solution do_install(const solution &s,
		      const a_iter &abegin, const a_iter &aend,
		      const forbid_iter &forbidden_iter)
  {
    int new_action_score=s.get_action_score();
    partial_solution tmpsol(s);
    std::set<dep> new_broken;

    for(a_iter a=abegin; a!=aend; ++a)
      {
	const version &v=a->ver;
	version old_version=s.version_of(v.get_package());

	assert(old_version == v.get_package().current_version() &&
	       old_version != v);

	new_action_score+=step_score;

	assert(v.get_id()<universe.get_version_count());
	new_action_score+=version_scores[v.get_id()];
	assert(v.get_package().current_version().get_id()<universe.get_version_count());
	new_action_score-=version_scores[v.get_package().current_version().get_id()];

	tmpsol.install(v.get_package(), v);
      }

    // Check all revdeps of the old AND new package versions, and all
    // previously broken deps.  Is there any way to speed this up?
    // For instance, by keeping a cache of already-examined deps.
    // Note that for the old version we only care about Depends and
    // for the new we only care about Conflicts...but it's not clear
    // that information can really be used in any good way (aside from
    // satisfied_by()).
    //
    // Note that this might be a rather expensive step -- on the other
    // hand, if lots of deps are broken, we might have problems
    // anyway.  Just plowing ahead for now.
    //
    // If you touch this code, remember that the revdeps list is not a
    // full list of the reverse dependencies; the constraint on it is
    // that any deps newly broken by installing v must appear in
    // *either* the reverse list of v *or* the reverse list of
    // old_version.


    // Note: I *think* it *might* be OK, if installing multiple
    // versions at once, to merge this loop with the above (think of
    // installing one version after another..); however, as I'm not
    // entirely certain and it's a relatively small optimization, I'm
    // taking the safe route and checking *all* versions.
    for(a_iter a=abegin; a!=aend; ++a)
      {
	const version &v=a->ver;
	version old_version=s.version_of(v.get_package());

	for(typename version::revdep_iterator rd=old_version.revdeps_begin();
	    !rd.end(); ++rd)
	  if((*rd).broken_under(tmpsol))
	    new_broken.insert(*rd);

	for(typename version::revdep_iterator rd=v.revdeps_begin();
	    !rd.end(); ++rd)
	  if((*rd).broken_under(tmpsol))
	    new_broken.insert(*rd);

	// Remember: the dependencies of the *new versions being
	// installed* might also be broken!
	for(typename version::dep_iterator di=v.deps_begin();
	    !di.end(); ++di)
	  if((*di).broken_under(tmpsol))
	    new_broken.insert(*di);
      }

    const std::set<dep> &old_broken=s.get_broken();
    for(typename std::set<dep>::const_iterator bd=old_broken.begin();
	bd!=old_broken.end(); ++bd)
      if((*bd).broken_under(tmpsol))
	new_broken.insert(*bd);

    int new_score=new_action_score+broken_score*new_broken.size();

    if(new_broken.size() == 0)
      new_score+=full_solution_score;

    return solution(abegin, aend, s,
		    new_broken,
		    forbidden_iter,
		    new_score,
		    new_action_score);
  }

  /** Enqueues a single successor node to the given solution by
   *  installing the given package version.
   *
   *  \param s the predecessor of the new solution
   *  \param v the version to install
   *  \param forbidden_iter an APT-style iterator over a set
   *  of versions that should be forbidden.
   *
   *  \return \b true if the new solution was not in the closed set.
   */
  template<class forbid_iter, class action_iter>
  bool try_install(const solution &s,
		   const action_iter &abegin, const action_iter &aend,
		   const forbid_iter &forbidden_iter)
  {
    solution s2=do_install(s, abegin, aend, forbidden_iter);

    if(!irrelevant(s2))
      {
	if(debug)
	  {
	    std::cout << "Enqueuing ";
	    s2.dump(std::cout);
	    std::cout << std::endl;
	  }

	open.push(s2);

	return true;
      }
    else
      {
	if(debug)
	  {
	    std::cout << "Dropping irrelevant solution ";
	    s2.dump(std::cout);
	    std::cout << std::endl;
	  }

	return false;
      }
  }

  /** Generates (and enqueues) successor nodes for the given broken
   *  dependency of the given solution.
   *
   *  \param s the solution to generate successors for
   *  \param d the dependency to fix
   *  \param count a value to increment for every successor that
   *         is enqueued or discarded (as long as it's discarded
   *         for being too bad, and not for being in closed).
   */
  void generate_successors(const solution &s,
			   const dep &d,
			   unsigned int &count)
  {
    version source_version=d.get_source();
    package source_package=source_version.get_package();
    version sol_version=s.version_of(source_package);

//     if(d.solvers_begin().end())
//       {
// 	std::cout << "No solvers for " << d.get_dep().ParentPkg().Name() 
// 		  << " " << d.get_dep().DepType() << " "
// 		  << d.get_dep().TargetPkg().Name();
// 	if(d.get_dep().TargetVer())
// 	  std::cout << " (" << d.get_dep().CompType()
// 		    << " " << d.get_dep().TargetVer() << ")";

// 	if(!d.get_prv().end())
// 	  std::cout << "  (through "
// 		    << d.get_prv().OwnerPkg().Name() << " "
// 		    << d.get_prv().OwnerVer().VerStr()
// 		    << " Provides "
// 		    << d.get_prv().ParentPkg().Name()
// 		    << ")";

// 	std::cout << std::endl;
//       }

    // if not, then the dep is satisfied!
    assert(sol_version == source_version);

    // If the source of the dependency was not modified and is
    // presently installed, try installing a *different version*.
    if(source_version == source_package.current_version())
      {
	for(typename package::version_iterator vi=source_package.versions_begin();
	    !vi.end(); ++vi)
	  if(*vi != source_version)
	    {
	      if(s.get_forbidden_versions().find(*vi) != s.get_forbidden_versions().end())
		{
		  if(debug)
		    std::cout << "Discarding forbidden version "
			      << (*vi).get_name() << std::endl;

		  continue;
		}

	      if(debug)
		std::cout << "  Trying to resolve " << d << " by installing " << (*vi).get_package().get_name() << " version " << (*vi).get_name() << std::endl;

	      action act(*vi);
	      if(try_install(s, &act, (&act)+1, d.solvers_begin()))
		++count;
	    }
      }

    // Enqueue each potential solver.
    //
    // (note: skip packages that have already been assigned...I coul
    // probably speed this up very slightly by pushing knowledge down
    // out of the abstract layer -- no need to iterate over all the
    // versions of each package that's already assigned)
    for(typename dep::solver_iterator si=d.solvers_begin();
	!si.end(); ++si)
      {
	version sol_ver=*si;

	if(sol_ver != sol_ver.get_package().current_version() &&
	   s.get_actions().find(sol_ver.get_package())==s.get_actions().end())
	  {
	    if(s.get_forbidden_versions().find(sol_ver) != s.get_forbidden_versions().end())
	      {
		if(debug)
		  std::cout << "Discarding forbidden version "
			    << sol_ver.get_name() << std::endl;

		continue;
	      }

	    if(debug)
	      std::cout << "  Trying to resolve " << d << " by installing " << (*si).get_package().get_name() << " version " << (*si).get_name() << std::endl;

	    action act(*si);
	    if(try_install(s, &act, (&act)+1, dummy_end_iterator<version>()))
	      ++count;
	  }
      }
  }

  /** Generate "forced" successors -- those successors which are
   *  logically required by the existing dependencies.
   *
   *  \return \b true iff the dependency can't be resolved at all
   *
   *  \param s the solution to try.
   *
   *  \param d the dep to test.
   *
   *  \param actions an out parameter -- any generated forced
   *  successor will be pushed onto this list.
   *
   *  \param toforbid an out parameter -- includes additional
   *  forbidden versions over what's forbidden by s.
   */
  bool generate_forced_successor(const solution &s,
				 const dep &d,
				 std::set<action> &actions,
				 std::set<version> &toforbid)
  {
    version v;
    version source_version=d.get_source();
    package source_package=source_version.get_package();
    bool found=false, from_source=false;

    if(debug)
      {
	std::cout << "Testing whether " << d << " is a forced dependency." << std::endl;
      }

    // If the source of the dependency was not modified and is
    // presently installed, try installing a *different version*.
    if(source_version == source_package.current_version())
      {
	for(typename package::version_iterator vi=source_package.versions_begin();
	    !vi.end(); ++vi)
	  // Note that testing against toforbid is strictly a
	  // performance hack (it would be caught in the next cycle if
	  // we didn't do it here, but why not use the information
	  // ASAP?)
	  if(*vi != source_version &&
	     s.get_forbidden_versions().find(*vi) == s.get_forbidden_versions().end() &&
	     toforbid.find(*vi)==toforbid.end())
	    {
	      if(found)
		return false;
	      found=true;
	      from_source=true;
	      v=*vi;
	    }
      }

    for(typename dep::solver_iterator si=d.solvers_begin();
	!si.end(); ++si)
      {
	version sol_ver=*si;

	if(sol_ver != sol_ver.get_package().current_version() &&
	   s.get_actions().find(sol_ver.get_package())==s.get_actions().end() &&
	   s.get_forbidden_versions().find(sol_ver) == s.get_forbidden_versions().end() &&
	   toforbid.find(sol_ver)==toforbid.end())
	  {
	    if(found)
	      return false;
	    found=true;
	    v=sol_ver;
	  }
      }

    if(found)
      {
	if(debug)
	    std::cout << "Forced resolution of " << d
		      << " by installing " << v.get_package().get_name()
		      << " version " << v.get_name() << std::endl;

	if(from_source)
	  for(typename dep::solver_iterator si=d.solvers_begin();
	      !si.end(); ++si)
	    toforbid.insert(*si);

	actions.insert(action(v));

	return false;
      }
    else
      {
	if(debug)
	  std::cout << "Discarding solution: unsolvable dependency " << d << std::endl;

	return true;
      }
  }

  /** Convert a [begin,end) pair to a single APT-style iterator.
   */
  template<class iter, class target>
  class apt_iter_wrapper
  {
    iter curr, the_end;
  public:
    apt_iter_wrapper(const iter &_begin, const iter &_end)
      :curr(_begin), the_end(_end)
    {
    }

    bool end() const
    {
      return curr==the_end;
    }

    apt_iter_wrapper &operator++()
    {
      ++curr;
      return *this;
    }

    const target &operator*() const
    {
      return *curr;
    }
  };

  /** Processes the given solution by enqueuing its successor nodes
   *  (if any are available).
   */
  void process_solution(const solution &s)
  {
    // How many solutions have been generated?
    unsigned int sols=0;

    std::set<action> forced_actions;
    std::set<version> forced_forbidden_versions;


    for(typename std::set<dep>::const_iterator bi=s.get_broken().begin();
	bi!=s.get_broken().end(); ++bi)
      if(generate_forced_successor(s, *bi,
				   forced_actions,
				   forced_forbidden_versions))
	return;

    if(!forced_actions.empty())
      {
	try_install(s,
		    forced_actions.begin(), forced_actions.end(),
		    apt_iter_wrapper<typename std::set<version>::const_iterator, version>(forced_forbidden_versions.begin(), forced_forbidden_versions.end()));
	return;
      }

    for(typename std::set<dep>::const_iterator bi=s.get_broken().begin();
	bi!=s.get_broken().end() && (sols==0 || sols<max_successors); ++bi)
      generate_successors(s, *bi, sols);
  }
public:

  /** Construct a new generic_problem_resolver.
   *
   *  \param _score_score the score per "step" of a (partial) solution.  Typically negative.
   *  \param _broken_score the score to add per broken dependency of a (partial) solution.  Typically negative.
   *  \param infinity a score value that will be considered to be "infinite".  Solutions
   *  with less than -infinity points will be immediately discarded.
   *  \param _full_solution_score a bonus for goal nodes (things
   *  that solve all dependencies)
   *  \param _universe the universe in which we are working.
   */
  generic_problem_resolver(int _step_score, int _broken_score,
			   int infinity, unsigned int _max_successors,
			   int _full_solution_score,
			   const PackageUniverse &_universe)
    :step_score(_step_score), broken_score(_broken_score),
     minimum_score(-infinity), max_successors(_max_successors),
     full_solution_score(_full_solution_score),
     universe(_universe), finished(false), debug(false)
  {
    version_scores=new int[universe.get_version_count()];
    for(size_t i=0; i<universe.get_version_count(); ++i)
      version_scores[i]=0;
  }

  ~generic_problem_resolver()
  {
    delete[] version_scores;
  }

  const PackageUniverse &get_universe()
  {
    return universe;
  }

  int get_step_score() {return step_score;}
  int get_broken_score() {return broken_score;}
  int get_infinity() {return -minimum_score;}
  int get_max_successors() {return max_successors;}
  int get_full_solution_score() {return full_solution_score;}

  /** Enables or disables debugging.  Debugging is initially
   *  disabled.
   */
  void set_debug(bool new_debug)
  {
    debug=new_debug;
  }

  /** Clears all the internal state of the solver, discards solutions,
   *  zeroes out scores.  Call this routine after changing the state
   *  of packages to avoid inconsistent results.
   */
  void reset()
  {
    finished=false;
    open.clear();
    closed.clear();

    for(size_t i=0; i<universe.get_version_count(); ++i)
      version_scores[i]=0;
  }

  /** \return \b true if no solutions have been examined yet.
   *  This implies that it is safe to modify package scores.
   */
  bool fresh() const
  {
    return open.empty() && !finished;
  }

  /** \return \b true if the open queue is empty. */
  bool exhausted() const
  {
    return open.empty() && finished;
  }

  /** Tells the resolver how highly to value a particular package
   *  version.  All scores are relative, and a higher score will
   *  result in a bias towards that version appearing in the final
   *  solution.
   */
  void set_version_score(const version &ver, int score)
  {
    assert(ver.get_id()<universe.get_version_count());
    version_scores[ver.get_id()]=score;
  }

  /** As set_version_score, but instead of replacing the current score
   *  increment it.
   */
  void add_version_score(const version &ver, int score)
  {
    assert(ver.get_id()<universe.get_version_count());
    version_scores[ver.get_id()]+=score;
  }

  /** \return the score of the version ver. */
  int get_version_score(const version &ver)
  {
    assert(ver.get_id()<universe.get_version_count());
    return version_scores[ver.get_id()];
  }

  /** Try to find the "next" solution: remove partial solutions from
   *  the open queue and place them in the closed queue until one of
   *  the following occurs:
   *
   *   - The number of broken dependencies drops to 0, in which case
   *     there is much rejoicing and we return successfully.
   *
   *   - The upper limit on the number of steps to perform is exceeded,
   *     in which case we just give up and report failure.  (this is a
   *     guard against exponential blowup)
   *
   *   - We run out of potential solutions to try; failure.
   *
   *  \param max_steps the maximum number of solutions to test.
   *
   *  \return a solution that fixes all broken dependencies
   *
   * \throws NoMoreSolutions if the potential solution list is exhausted.
   * \throws NoMoreTime if no solution is found within max_steps steps.
   *
   *  \todo when throwing NoMoreSolutions or NoMoreTime, maybe we
   *        should include the "least broken" solution seen.
   */
  solution find_next_solution(int max_steps)
  {
    if(finished)
      throw NoMoreSolutions();

    // If the open queue is empty, then we're between searches and
    // should enqueue a new root node.
    if(open.empty())
      {
	closed.clear();

	std::set<dep> broken;

	// Find all the broken deps.
	for(typename PackageUniverse::broken_dep_iterator bi=universe.broken_begin();
	    !bi.end(); ++bi)
	  broken.insert(*bi);

	open.push(solution(broken, broken.size()*broken_score));
      }

    while(max_steps>0 && !open.empty())
      {
	solution s=open.top();
	open.pop();

	if(irrelevant(s))
	  {
	    if(debug)
	      {
		std::cout << "Dropping irrelevant solution ";
		s.dump(std::cout);
		std::cout << std::endl;
	      }
	    continue;
	  }

	closed.insert(s);

	if(debug)
	  {
	    std::cout << "Processing ";
	    s.dump(std::cout);
	    std::cout << std::endl;
	  }

	// If all dependencies are satisfied, we found a solution.
	if(s.get_broken().empty())
	  {
	    if(debug)
	      {
		std::cout << " --- Found solution ";
		s.dump(std::cout);
		std::cout << std::endl;
	      }

	    // Set it here too in case the last tentative solution is
	    // really a solution.
	    if(open.empty())
	      finished=true;

	    generated_solutions.push_back(s);

	    return s;
	  }
	// Nope, let's go enqueue successor nodes.
	else
	  process_solution(s);

	if(debug)
	  std::cout << "Done generating successors." << std::endl;

	--max_steps;
      }

    // Oh no, we either ran out of solutions or ran out of steps.

    if(max_steps==0)
      throw NoMoreTime();

    assert(open.empty());

    finished=true;
    throw NoMoreSolutions();
  }

  void dump_scores(std::ostream &out)
  {
    out << "{" << std::endl;
    for(typename PackageUniverse::package_iterator i=universe.packages_begin();
	!i.end(); ++i)
      {
	bool any_modified=false;

	for(typename PackageUniverse::package::version_iterator j=(*i).versions_begin();
	    !j.end(); ++j)
	  if(version_scores[(*j).get_id()]!=0)
	    any_modified=true;

	if(any_modified)
	  {
	    out << "  SCORE " << (*i).get_name() << " <";

	    for(typename PackageUniverse::package::version_iterator j=(*i).versions_begin();
	    !j.end(); ++j)
	      if(version_scores[(*j).get_id()]!=0)
		out << " " << (*j).get_name() << " " << version_scores[(*j).get_id()];

	    out << " >" << std::endl;
	  }
      }
    out << "}" << std::endl;
  }
};


template<class PackageUniverse>
void dump_universe(const PackageUniverse &world, std::ostream &out)
{
  out << "UNIVERSE [";
  for(typename PackageUniverse::package_iterator p=world.packages_begin();
      !p.end(); ++p)
    {
      out << "  PACKAGE " << (*p).get_name() << " < ";
      for(typename PackageUniverse::package::version_iterator v=(*p).versions_begin();
	  !v.end(); ++v)
	out << (*v).get_name() << " ";
      out << ">" << " " << (*p).current_version().get_name() << std::endl;
    }

  for(typename PackageUniverse::dep_iterator d=world.deps_begin();
      !d.end(); ++d)
    {
      const typename PackageUniverse::version &sv=(*d).get_source();
      const typename PackageUniverse::package &sp=sv.get_package();

      out << "  DEP " << sp.get_name() << " " << sv.get_name() << " -> < ";

      for(typename PackageUniverse::dep::solver_iterator t=(*d).solvers_begin();
	  !t.end(); ++t)
	{
	  out << " " << (*t).get_package().get_name() << " " << (*t).get_name() << " ";
	}
      out << " > " << std::endl;
    }
  out << "]" << std::endl;
}

#endif
