// solution_dialog.cc
//
//   Copyright (C) 2005 Daniel Burrows
//
//   This program is free software; you can redistribute it and/or
//   modify it under the terms of the GNU General Public License as
//   published by the Free Software Foundation; either version 2 of
//   the License, or (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//   General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; see the file COPYING.  If not, write to
//   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA 02111-1307, USA.

#include "solution_dialog.h"
#include "solution_fragment.h"
#include "ui.h"

#include <aptitude.h>

#include <vscreen/config/colors.h>
#include <vscreen/config/keybindings.h>
#include <vscreen/fragment.h>
#include <vscreen/vs_button.h>
#include <vscreen/vs_center.h>
#include <vscreen/vs_frame.h>
#include <vscreen/vs_scrollbar.h>
#include <vscreen/vs_table.h>
#include <vscreen/vs_text_layout.h>

#include <generic/apt.h>
#include <generic/aptitude_resolver.h>

#include <sigc++/bind.h>

class solution_dialog:public vs_text_layout
{
  aptitude_resolver::solution last_sol;

  void handle_cache_reload()
  {
    if(apt_cache_file)
      {
	(*apt_cache_file)->package_state_changed.connect(sigc::mem_fun(*this, &solution_dialog::update));
	(*apt_cache_file)->selected_solution_changed.connect(sigc::mem_fun(*this, &solution_dialog::update));
      }

    update();
  }
public:
  solution_dialog()
  {
    if(apt_cache_file)
      {
	(*apt_cache_file)->package_state_changed.connect(sigc::mem_fun(*this, &solution_dialog::update));
	(*apt_cache_file)->selected_solution_changed.connect(sigc::mem_fun(*this, &solution_dialog::update));
      }

    cache_closed.connect(sigc::mem_fun(*this, &solution_dialog::update));
    cache_reloaded.connect(sigc::mem_fun(*this, &solution_dialog::handle_cache_reload));

    update();
  }

  void update()
  {
    if(!apt_cache_file)
      {
	set_fragment(fragf("%s", _("The package cache is not available.")));
	last_sol.nullify();
	return;
      }

    if(!(*apt_cache_file)->resolver_exists())
      {
	// This makes ASS-U-MPTIONS about how resolver_exists works.
	set_fragment(fragf("%s", _("No packages are broken.")));
	last_sol.nullify();
	return;
      }

    try
      {
	aptitude_resolver::solution sol=(*apt_cache_file)->get_current_solution();

	if(sol == last_sol)
	  return;

	last_sol=sol;

	if(sol.get_actions().empty())
	  set_fragment(fragf("%s", _("Internal error: unexpected null solution.")));
	else
	  set_fragment(solution_fragment(sol));
      }
    catch(NoMoreSolutions)
      {
	set_fragment(fragf("%s", _("No resolution found.")));
      }
    catch(NoMoreTime)
      {
	set_fragment(fragf(_("Time exhausted while searching for a solution (you can select \"Next Solution\" or press %s to try harder)."),
			   global_bindings.readable_keyname("NextSolution").c_str()));
      }
  }
};

static void do_apply(vscreen_widget *w)
{
  do_apply_solution();
  w->destroy();
}

vscreen_widget *make_solution_dialog()
{
  vs_table *t=new vs_table;
  vscreen_widget *rval=new vs_center(new vs_frame(t));

  vs_text_layout *display=new solution_dialog;
  vs_scrollbar *scrl=new vs_scrollbar(vs_scrollbar::VERTICAL);

  display->location_changed.connect(sigc::mem_fun(scrl, &vs_scrollbar::set_slider));
  scrl->scrollbar_interaction.connect(sigc::mem_fun(display, &vs_text_layout::scroll));

  t->add_widget_opts(display,
		     0, 0, 1, 1,
		     vs_table::EXPAND | vs_table::FILL | vs_table::SHRINK,
		     vs_table::EXPAND | vs_table::FILL | vs_table::SHRINK);

  t->add_widget_opts(scrl,
		     0, 1, 1, 1,
		     vs_table::ALIGN_RIGHT,
		     vs_table::ALIGN_CENTER | vs_table::FILL | vs_table::SHRINK);

  vs_table *bt=new vs_table;

  //t->set_bg(get_color("DefaultWidgetBackground")|A_REVERSE);

  // TODO: for dialogs like this, I really should have support for
  // "wrapping" lines of buttons if they get too long, like fragments.
  vs_button *bprev=new vs_button(_("Previous"));
  vs_button *bnext=new vs_button(_("Next"));
  vs_button *bapply=new vs_button(_("Apply"));
  vs_button *bclose=new vs_button(_("Close"));

  bprev->pressed.connect(sigc::ptr_fun(do_previous_solution));
  bnext->pressed.connect(sigc::ptr_fun(do_next_solution));
  bapply->pressed.connect(sigc::bind(sigc::ptr_fun(do_apply),
				     rval));
  bclose->pressed.connect(sigc::mem_fun(rval,
					&vscreen_widget::destroy));

  rval->connect_key("ApplySolution", &global_bindings,
		    sigc::bind(sigc::ptr_fun(do_apply),
			       rval));

  bprev->set_bg_style(style_attrs_flip(A_REVERSE));
  bnext->set_bg_style(style_attrs_flip(A_REVERSE));
  bapply->set_bg_style(style_attrs_flip(A_REVERSE));
  bclose->set_bg_style(style_attrs_flip(A_REVERSE));

  bt->add_widget_opts(bprev,
		     0, 0, 1, 1,
		     vs_table::FILL | vs_table::EXPAND, vs_table::FILL);
  bt->add_widget_opts(bnext,
		     0, 1, 1, 1,
		     vs_table::FILL | vs_table::EXPAND, vs_table::FILL);
  bt->add_widget_opts(bapply,
		     0, 2, 1, 1,
		     vs_table::FILL | vs_table::EXPAND, vs_table::FILL);
  bt->add_widget_opts(bclose,
		     0, 3, 1, 1,
		     vs_table::FILL | vs_table::EXPAND, vs_table::FILL);

  t->add_widget_opts(bt, 1, 0, 1, 2,
		     vs_table::FILL | vs_table::EXPAND | vs_table::SHRINK,
		     vs_table::FILL);

  return rval;
}
