// solution_fragment.cc
//
//
//   Copyright (C) 2005 Daniel Burrows
//
//   This program is free software; you can redistribute it and/or
//   modify it under the terms of the GNU General Public License as
//   published by the Free Software Foundation; either version 2 of
//   the License, or (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//   General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; see the file COPYING.  If not, write to
//   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA 02111-1307, USA.

#include "solution_fragment.h"

#include <aptitude.h>

#include <generic/aptitude_resolver.h>

#include <vscreen/fragment.h>

#include <vector>

using namespace std;

struct pkg_name_lt
{
public:
  bool operator()(const pkgCache::PkgIterator &p1,
		  const pkgCache::PkgIterator &p2) const
  {
    return strcmp(p1.Name(), p2.Name()) < 0;
  }
};

struct ver_name_lt
{
public:
  bool operator()(const pkgCache::VerIterator &v1,
		  const pkgCache::VerIterator &v2) const
  {
    return strcmp(v1.ParentPkg().Name(), v2.ParentPkg().Name()) < 0;
  }
};

/** \return a fragment listing the archives to which a version
 *  belongs in the form "archive1,archive2,..."
 */
fragment *archives_fragment(const pkgCache::VerIterator &v)
{
  vector<fragment *> archive_fragments;

  for(pkgCache::VerFileIterator vf=v.FileList(); !vf.end(); ++vf)
    {
      if(vf.File().Archive())
	archive_fragments.push_back(text_fragment(vf.File().Archive()));
      else
	archive_fragments.push_back(text_fragment(_("<NULL>")));
    }

  return join_fragments(archive_fragments, L",");
}

fragment *solution_fragment(const aptitude_resolver::solution &sol)
{
  // Bin packages according to what will happen to them.
  vector<pkgCache::PkgIterator> remove_packages;
  vector<pkgCache::PkgIterator> keep_packages;
  vector<pkgCache::VerIterator> install_packages;
  vector<pkgCache::VerIterator> downgrade_packages;
  vector<pkgCache::VerIterator> upgrade_packages;

  sort(remove_packages.begin(), remove_packages.end(),
       pkg_name_lt());
  sort(keep_packages.begin(), keep_packages.end(),
       pkg_name_lt());
  sort(install_packages.begin(), install_packages.end(),
       ver_name_lt());
  sort(downgrade_packages.begin(), downgrade_packages.end(),
       ver_name_lt());
  sort(upgrade_packages.begin(), upgrade_packages.end(),
       ver_name_lt());

  for(std::map<aptitude_resolver::package,
	aptitude_resolver::action>::const_iterator i=sol.get_actions().begin();
      i!=sol.get_actions().end(); ++i)
    {
      pkgCache::PkgIterator pkg=i->first.get_pkg();
      pkgCache::VerIterator curver=pkg.CurrentVer();
      pkgCache::VerIterator newver=i->second.ver.get_ver();

      if(curver.end())
	{
	  if(newver.end())
	    keep_packages.push_back(pkg);
	  else
	    install_packages.push_back(newver);
	}
      else if(newver.end())
	remove_packages.push_back(pkg);
      else if(newver == curver)
	keep_packages.push_back(pkg);
      else
	{
	  int cmp=_system->VS->CmpVersion(curver.VerStr(),
					  newver.VerStr());

	  // The versions shouldn't be equal -- otherwise
	  // something is majorly wrong.
	  // assert(cmp!=0);
	  //
	  // The above is not true: consider, eg, the case of a
	  // locally compiled package and a standard package.

	  /** \todo indicate "sidegrades" separately? */
	  if(cmp<=0)
	    upgrade_packages.push_back(newver);
	  else if(cmp>0)
	    downgrade_packages.push_back(newver);
	}
    }

  vector<fragment *> fragments;

  if(!remove_packages.empty())
    {
      fragments.push_back(fragf(_("%BRemove%b the following packages:%n")));
      for(vector<pkgCache::PkgIterator>::const_iterator i=remove_packages.begin();
	  i!=remove_packages.end(); ++i)
	fragments.push_back(fragf("  %s%n", i->Name()));

      fragments.push_back(newline_fragment());
    }

  if(!install_packages.empty())
    {
      fragments.push_back(fragf(_("%BInstall%b the following packages:%n")));
      for(vector<pkgCache::VerIterator>::const_iterator i=install_packages.begin();
	  i!=install_packages.end(); ++i)
	fragments.push_back(fragf("  %s [%s (%F)]%n",
				  i->ParentPkg().Name(),
				  i->VerStr(),
				  archives_fragment(*i)));

      fragments.push_back(newline_fragment());
    }

  if(!keep_packages.empty())
    {
      fragments.push_back(fragf(_("%BKeep%b the following packages at their current version:%n")));
      for(vector<pkgCache::PkgIterator>::const_iterator i=keep_packages.begin();
	  i!=keep_packages.end(); ++i)
	{
	  if(i->CurrentVer().end())
	    fragments.push_back(fragf("  %s [%s]%n",
				      i->Name(),
				      _("Not Installed")));
	  else
	    fragments.push_back(fragf("  %s [%s (%F)]%n",
				      i->Name(),
				      i->CurrentVer().VerStr(),
				      archives_fragment(i->CurrentVer())));
	}

      fragments.push_back(newline_fragment());
    }

  if(!upgrade_packages.empty())
    {
      fragments.push_back(fragf(_("%BUpgrade%b the following packages:%n")));
      for(vector<pkgCache::VerIterator>::const_iterator i=upgrade_packages.begin();
	  i!=upgrade_packages.end(); ++i)
	fragments.push_back(fragf("  %s [%s (%F) -> %s (%F)]%n",
				  i->ParentPkg().Name(),
				  i->ParentPkg().CurrentVer().VerStr(),
				  archives_fragment(i->ParentPkg().CurrentVer()),
				  i->VerStr(),
				  archives_fragment(*i)));

      fragments.push_back(newline_fragment());
    }

  if(!downgrade_packages.empty())
    {
      fragments.push_back(fragf(_("%BDowngrade%b the following packages:%n")));
      for(vector<pkgCache::VerIterator>::const_iterator i=downgrade_packages.begin();
	  i!=downgrade_packages.end(); ++i)
	fragments.push_back(fragf("  %s [%s (%F) -> %s (%F)]%n",
				  i->ParentPkg().Name(),
				  i->ParentPkg().CurrentVer().VerStr(),
				  archives_fragment(i->ParentPkg().CurrentVer()),
				  i->VerStr(),
				  archives_fragment(*i)));

      fragments.push_back(newline_fragment());
    }

  char buf[512];
  snprintf(buf, 512, _("Score is %d"), sol.get_score());
  fragments.push_back(fragf("%s", buf));

  return flowbox(sequence_fragment(fragments));
}
