// vs_menubar.cc

#include "vs_container.h"
#include "vs_menubar.h"
#include "vscreen.h"

#include "config/colors.h"

#include <sigc++/adaptors/bind.h>
#include <sigc++/functors/mem_fun.h>

using namespace std;

keybindings *vs_menubar::bindings=NULL;

vs_menubar::vs_menubar(bool _always_visible)
  :vs_container(), active(false), always_visible(_always_visible), curloc(0), subwidget(NULL)
{
  do_layout.connect(sigc::mem_fun(*this, &vs_menubar::layout_me));

  focussed.connect(sigc::mem_fun(*this, &vs_menubar::got_focus));
  unfocussed.connect(sigc::mem_fun(*this, &vs_menubar::lost_focus));
}

vs_menubar::~vs_menubar()
{
}

void vs_menubar::got_focus()
{
  vscreen_widget *w=get_focus();
  if(w)
    w->focussed();
}

void vs_menubar::lost_focus()
{
  vscreen_widget *w=get_focus();
  if(w)
    w->unfocussed();
}

bool vs_menubar::focus_me()
{
  if(active)
    return true;
  else if(subwidget && subwidget->focus_me())
    return true;
  else
    return vscreen_widget::focus_me();
}

vscreen_widget *vs_menubar::get_focus()
{
  if(active)
    {
      if(!active_menus.empty())
	return active_menus.front();
      else
	return NULL;
    }
  else if(subwidget)
    return subwidget;
  else
    return NULL;
}

bool vs_menubar::get_cursorvisible()
{
  vscreen_widget *w=get_focus();
  return (w && w->get_cursorvisible()) || (!w && active);
}

point vs_menubar::get_cursorloc()
{
  vscreen_widget *w=get_focus();

  if(w)
    {
      point p=w->get_cursorloc();
      p.x+=w->get_startx();
      p.y+=w->get_starty();
      return p;
    }
  else if(active)
    return point(get_menustart(curloc), 0);
  else
    return point(0, 0);
}

int vs_menubar::get_menustart(int idx)
{
  int rval=0;
  for(int i=0; i<idx; i++)
    {
      const wstring &title=items[i].title;
      rval+=wcswidth(title.c_str(), title.size());
    }
  return rval;
}

void vs_menubar::append_item(wstring title, vs_menu *menu)
{
  items.push_back(item(L' '+title+L' ', menu));

  menu->shown_sig.connect(sigc::bind(sigc::mem_fun(*this, &vs_menubar::show_menu), menu));
  menu->hidden_sig.connect(sigc::bind(sigc::mem_fun(*this, &vs_menubar::hide_menu), menu));
  menu->menus_goaway.connect(sigc::mem_fun(*this, &vs_menubar::disappear));
  menu->set_owner(this);

  vscreen_update();
}

void vs_menubar::set_subwidget(vscreen_widget *w)
{
  if(subwidget)
    subwidget->set_owner(NULL);

  if(subwidget)
    subwidget->unfocussed();

  subwidget=w;

  if(subwidget)
    subwidget->set_owner(this);

  if(subwidget)
    subwidget->focussed();

  vscreen_queuelayout();
}

void vs_menubar::show_all()
{
  if(subwidget)
    subwidget->show_all();
}

int vs_menubar::width_request()
{
  int w=0;

  // Calculate the size of the bar itself.
  for(itemlist::size_type i=0; i<items.size(); i++)
    {
      const wstring &title=items[i].title;

      w+=wcswidth(title.c_str(), title.size());
    }

  // Expand the width as needed to account for active menus.
  int num=0;
  for(activemenulist::iterator i=active_menus.begin();
      i!=active_menus.end();
      i++, num++)
    {
      int menux=0;

      // Calculate the starting X location of this menu.
      for(itemlist::size_type j=0; j<items.size(); j++)
	{
	  if(items[j].menu==*i)
	    break;

	  const wstring &title=items[j].title;

	  menux+=wcswidth(title.c_str(), title.size());
	}

      // Now expand our width request.
      w=max(w, menux+(*i)->width_request());
    }

  // Expand the width to account for the subwidget.
  if(subwidget)
    w=max(w, subwidget->width_request());

  return w;
}

// TODO: What if the width is insufficient: should I scroll left/right
// or wrap to the next line?
int vs_menubar::height_request(int w)
{
  int h=always_visible?1:0;

  int num=0;
  for(activemenulist::iterator i=active_menus.begin();
      i!=active_menus.end();
      i++, num++)
    h=max(h, 1+(*i)->height_request(w));

  if(subwidget)
    {
      int subwidget_h=subwidget->height_request(w);

      if(always_visible)
	subwidget_h+=1;

      h=max(h, subwidget_h);
    }

  return h;
}

void vs_menubar::layout_me()
{
  // Find the starting X location of each active menu.
  for(activemenulist::iterator i=active_menus.begin();
      i!=active_menus.end();
      i++)
    {
      int menux=0;

      for(itemlist::size_type j=0; j<items.size(); j++)
	{
	  if(items[j].menu==*i)
	    break;

	  const wstring &title=items[j].title;

	  menux+=wcswidth(title.c_str(), title.size());
	}

      int req_w=(*i)->width_request();
      if(getmaxx()<menux+req_w)
	{
	  if(getmaxx()>=req_w)
	    menux=getmaxx()-req_w;
	  else
	    {
	      menux=0;
	      req_w=getmaxx();
	    }
	}

      int req_h=(*i)->height_request(req_w);

      if(getmaxy()<1+req_h)
	req_h=getmaxy()-1;

      (*i)->alloc_size(menux,
		       1,
		       req_w,
		       req_h);
    }

  if(subwidget)
    subwidget->alloc_size(0,
			  always_visible?1:0,
			  getmaxx(),
			  always_visible?getmaxy()-1:getmaxy());
}

void vs_menubar::show_menu(vs_menu *w)
{
  if(active)
    {
      vscreen_widget *old_focus=get_focus();

      for(activemenulist::iterator i=active_menus.begin();
	  i!=active_menus.end();
	  i++)
	assert(w!=*i);

      if(old_focus)
	old_focus->unfocussed();

      active_menus.push_front(w);

      w->focussed();

      vscreen_queuelayout();
      vscreen_update();
    }
}

void vs_menubar::hide_menu(vs_menu *w)
{
  if(active)
    {
      for(activemenulist::iterator i=active_menus.begin();
	  i!=active_menus.end();
	  i++)
	{
	  if(*i==w)
	    {
	      w->unfocussed();
	      active_menus.remove(w);

	      vscreen_widget *new_focus=get_focus();
	      if(new_focus)
		new_focus->focussed();

	      vscreen_queuelayout();
	      vscreen_update();
	      return;
	    }
	}

      abort();
    }
}

void vs_menubar::appear()
{
  if(!active)
    {
      active=true;
      if(subwidget)
	subwidget->unfocussed();

      // Added to eliminate the weird "titles are selected but contents aren't"
      // state that was normal with the previous settings:
      if(items.size()>0)
	items[curloc].menu->show();

      vscreen_update();
    }
}

void vs_menubar::disappear()
{
  if(active)
    {
      while(!active_menus.empty())
	active_menus.front()->hide();

      active=false;
      if(subwidget)
	subwidget->focussed();

      curloc=0;

      vscreen_update();
    }
}

bool vs_menubar::handle_key(const key &k)
{
  if(bindings->key_matches(k, "ToggleMenuActive"))
    {
      if(active)
	disappear();
      else
	appear();
    }
  else if(active)
    {
      if(bindings->key_matches(k, "Cancel"))
	{
	  disappear();

	  //if(!active_menus.empty())
	  //  active_menus.front()->hide();
	  //else
	  //  disappear();

	  vscreen_update();
	}
      else if(!active_menus.empty())
	{
	  if(bindings->key_matches(k, "Right"))
	    {
	      if(items.size()>0)
		{
		  while(!active_menus.empty())
		    active_menus.front()->hide();

		  active_menus.clear();

		  if(curloc<items.size()-1)
		    curloc++;
		  else
		    curloc=0;

		  items[curloc].menu->show();

		  vscreen_update();
		}
	    }
	  else if(bindings->key_matches(k, "Left"))
	    {
	      if(items.size()>0)
		{
		  while(!active_menus.empty())
		    active_menus.front()->hide();

		  active_menus.clear();

		  if(curloc>0)
		    curloc--;
		  else
		    curloc=items.size()-1;

		  items[curloc].menu->show();
		  vscreen_update();
		}
	    }
	  else if(active_menus.front()->dispatch_key(k))
	    return true;
	  else
	    return vscreen_widget::handle_key(k);
	}
      else if(bindings->key_matches(k, "Right"))
	{
	  if(items.size()>0)
	    {
	      if(curloc<items.size()-1)
		curloc++;
	      else
		curloc=0;

	      vscreen_update();
	    }
	}
      else if(bindings->key_matches(k, "Left"))
	{
	  if(items.size()>0)
	    {
	      if(curloc>0)
		curloc--;
	      else
		curloc=items.size()-1;

	      vscreen_update();
	    }
	}
      else if(bindings->key_matches(k, "Down") ||
	      bindings->key_matches(k, "Confirm"))
	{
	  if(items.size()>0)
	    items[curloc].menu->show();
	}
      else
	return vscreen_widget::handle_key(k);

      return true;
    }
  else if(subwidget && subwidget->dispatch_key(k))
    return true;
  else
    return vscreen_widget::handle_key(k);

  return true;
}

void vs_menubar::paint(const style &st)
{
  if(subwidget)
    subwidget->display(st);

  if(active || always_visible)
    {
      const style menubar_style=get_style("MenuBar");
      const style highlightedmenubar_style=get_style("HighlightedMenuBar");

      if(active)
	for(activemenulist::reverse_iterator i=active_menus.rbegin();
	    i!=active_menus.rend();
	    i++)
	  (*i)->display(st);

      int pos=0, maxx=getmaxx();

      apply_style(menubar_style);
      move(0, 0);
      for(int i=0; i<maxx; i+=wcwidth(L' '))
	add_wch(L' ');

      move(0, 0);
      for(itemlist::size_type i=0; i<items.size() && pos<maxx; i++)
	{
	  if(active && i==curloc)
	    apply_style(highlightedmenubar_style);
	  else
	    apply_style(menubar_style);

	  wstring &title = items[i].title;
	  size_t titleloc = 0;

	  while(titleloc < title.size() && pos < maxx)
	    {
	      wchar_t wch=title[titleloc];

	      add_wch(wch);
	      pos+=wcwidth(wch);
	      ++titleloc;
	    }
	}
    }
}

void vs_menubar::dispatch_mouse(short id, int x, int y, int z, mmask_t bmask)
{
  if(y==0)
    {
      if(bmask & (BUTTON1_CLICKED | BUTTON2_CLICKED |
		  BUTTON3_CLICKED | BUTTON4_CLICKED |
		  BUTTON1_RELEASED | BUTTON2_RELEASED |
		  BUTTON3_RELEASED | BUTTON4_RELEASED |
		  BUTTON1_PRESSED | BUTTON2_PRESSED |
		  BUTTON3_PRESSED | BUTTON4_PRESSED))
	{
	  if(!active)
	    appear();

	  int loc=0;
	  activemenulist::size_type i=0;

	  if(items.size()>0)
	    {
	      loc+=items[0].title.size();

	      while(i<items.size()-1 && loc<=x)
		{
		  loc+=items[i+1].title.size();

		  ++i;
		}
	    }

	  if(i<items.size())
	    {
	      while(!active_menus.empty())
		active_menus.front()->hide();

	      active_menus.clear();

	      curloc=i;

	      items[curloc].menu->show();

	      vscreen_update();
	    }
	}
    }
  else if(active)
    {
      for(activemenulist::iterator i=active_menus.begin();
	  i!=active_menus.end();
	  i++)
	if((*i)->enclose(y, x))
	  {
	    (*i)->dispatch_mouse(id,
				 x-(*i)->get_startx(), y-(*i)->get_starty(), z,
				 bmask);
	    return;
	  }
    }

  if(subwidget)
    subwidget->dispatch_mouse(id,
			      x-subwidget->get_startx(),
			      y-subwidget->get_starty(), z, bmask);
}

void vs_menubar::init_bindings()
{
  bindings=new keybindings(&global_bindings);
}

void vs_menubar::set_always_visible(bool _always_visible)
{
  if(_always_visible!=always_visible)
    {
      always_visible=_always_visible;
      vscreen_update();
      vscreen_queuelayout();
    }
}
