// vscreen.cc
//
//  Copyright 1999-2005 Daniel Burrows
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; see the file COPYING.  If not, write to
//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//  Boston, MA 02111-1307, USA.
//
//  Implementation of vscreen stuff.

// We need this for recursive mutexes.  Should all files be compiled
// with this defined??
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include "vscreen.h"
#include "curses++.h"
#include "vscreen_widget.h"

#include "transcode.h"
#include "vs_editline.h"
#include "vs_menu.h"
#include "vs_menubar.h"
#include "vs_pager.h"
#include "vs_statuschoice.h"
#include "vs_table.h"
#include "vs_text_layout.h"
#include "vs_tree.h"

#include "config/keybindings.h"
#include "config/style.h"

// For _()
#include "../aptitude.h"

#include <signal.h>

#ifdef HAVE_LIBPTHREAD
#include <pthread.h>
#endif

#include <assert.h>
#include <sys/time.h>

#ifdef HAVE_LIBPTHREAD
static pthread_mutex_t vscreen_mutex;
#endif

using namespace std;

bool curses_avail=false;
bool should_exit=false;

// Used to queue and merge update requests
static bool needs_layout=false;
static bool needs_update=false;
static bool needs_cursorupdate=false;

sigc::signal0<void> main_hook;

struct timeout_info
{
  sigc::slot0<void> f;
  int num;
  timeval activate_time; // tells when this timeout should be triggered
  timeout_info(const sigc::slot0<void> &_f, int _num, timeval &_activate_time)
    :f(_f), num(_num), activate_time(_activate_time)
  {}
};

list<timeout_info> timeouts;
// Timeouts which have been registered.  For optimal speed this should be a
// heap, but there are other considerations...it might be necessary for callers
// to delete arbitrary timeouts, and while maintaining heap order with
// arbitrary deletions might be possible, it's a bit of a nuisance.  Also,
// I don't expect this to ever hold more than a handful of values.
//  In any event, the external interface doesn't depend on details like
// this, so it will be fairly simple to switch to using a more efficient
// data structure internally if this becomes a problem.  I notice that glib
// uses something similar to this, and it's efficient enough..  Optimizing
// stuff that doesn't need to be optimized is a waste of time.
int ntimeouts=0;
// Cheesy way of generating mostly-unique IDs for timeouts -- just incremented 
// once for every timeout that gets added.

static vscreen_widget *toplevel=NULL;
// The widget which is displayed as the root of everything

//static bool resized=false;
static void sigresize(int sig)
{
  vscreen_handleresize();
  //resized=true;
}

// Cleanly shutdown (eg, restore screen settings if possible)
//
// Called on SIGTERM, SIGINT, SIGSEGV, SIGABRT, and SIGQUIT
//
// FIXME: revert to the /previous/ handler, not just SIG_DFL?
static void sigkilled(int sig)
{
  endwin();

  switch(sig)
    {
    case SIGTERM:
      fprintf(stderr, _("Ouch!  Got SIGTERM, dying..\n"));
      break;
    case SIGSEGV:
      fprintf(stderr, _("Ouch!  Got SIGSEGV, dying..\n"));
      break;
    case SIGABRT:
      fprintf(stderr, _("Ouch!  Got SIGABRT, dying..\n"));
      break;
    case SIGQUIT:
      fprintf(stderr, _("Ouch!  Got SIGQUIT, dying..\n"));
      break;
    }

  signal(sig, SIG_DFL);
  raise(sig);
}

///////////////////////////////////////////////////////////////////////////////
// The following function comes from the glibc documentation.
// Subtract the `struct timeval' values X and Y,
// storing the result in RESULT.
// Return 1 if the difference is negative, otherwise 0.

static int
timeval_subtract (timeval *result, timeval *x, timeval *y)
{
  /* Perform the carry for the later subtraction by updating Y. */
  if (x->tv_usec < y->tv_usec)
    {
      int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
      y->tv_usec -= 1000000 * nsec;
      y->tv_sec += nsec;
    }
  if (x->tv_usec - y->tv_usec > 1000000)
    {
      int nsec = (x->tv_usec - y->tv_usec) / 1000000;
      y->tv_usec += 1000000 * nsec;
      y->tv_sec -= nsec;
    }

  /* Compute the time remaining to wait.
     `tv_usec' is certainly positive. */
  result->tv_sec = x->tv_sec - y->tv_sec;
  result->tv_usec = x->tv_usec - y->tv_usec;

  /* Return 1 if result is negative. */
  return x->tv_sec < y->tv_sec;
}
///////////////////////////////////////////////////////////////////////////////

void vscreen_settoplevel(vscreen_widget *w)
{
  if(toplevel)
    toplevel->unfocussed();

  toplevel=w;

  if(curses_avail)
    {
      toplevel->set_owner_window(rootwin, 0, 0, rootwin.getmaxx(), rootwin.getmaxy());
      toplevel->show_all();
      toplevel->focussed();
      vscreen_redraw();
    }
}

void vscreen_init()
{
  keybinding upkey, downkey, leftkey, rightkey, quitkey, homekey, endkey;
  keybinding historynextkey, historyprevkey;
  keybinding delfkey, delbkey, ppagekey, npagekey;
  keybinding undokey, helpkey, researchkey;
  keybinding menutogglekey, cancelkey;

  upkey.push_back(key(KEY_UP, true));
  upkey.push_back(key(L'k', false));
  downkey.push_back(key(KEY_DOWN, true));
  downkey.push_back(key(L'j', false));
  leftkey.push_back(key(KEY_LEFT, true));
  leftkey.push_back(key(L'h', false));
  rightkey.push_back(key(KEY_RIGHT, true));
  rightkey.push_back(key(L'l', false));
  quitkey.push_back(key(L'q', false));

  historyprevkey.push_back(key(KEY_UP, true));
  historyprevkey.push_back(KEY_CTRL(L'p'));
  historynextkey.push_back(key(KEY_DOWN, true));
  historynextkey.push_back(KEY_CTRL(L'n'));

  homekey.push_back(key(KEY_HOME, true));
  homekey.push_back(KEY_CTRL(L'a'));
  endkey.push_back(key(KEY_END, true));
  endkey.push_back(KEY_CTRL(L'e'));

  delfkey.push_back(key(KEY_DC, true));
  delfkey.push_back(KEY_CTRL(L'd'));

  delbkey.push_back(key(KEY_BACKSPACE, true));
  delbkey.push_back(KEY_CTRL(L'h'));

  ppagekey.push_back(key(KEY_PPAGE, true));
  ppagekey.push_back(KEY_CTRL(L'b'));

  npagekey.push_back(key(KEY_NPAGE, true));
  npagekey.push_back(KEY_CTRL(L'f'));

  undokey.push_back(KEY_CTRL(L'u'));
  undokey.push_back(KEY_CTRL(L'_'));

  helpkey.push_back(key(L'?', false));
  helpkey.push_back(KEY_CTRL(L'h'));
  helpkey.push_back(key(KEY_F(1), true));

  menutogglekey.push_back(key(KEY_F(10), true));
  menutogglekey.push_back(KEY_CTRL(L' '));

  cancelkey.push_back(KEY_CTRL(L'g'));
  cancelkey.push_back(key(L'\e', true));
  cancelkey.push_back(KEY_CTRL(L'['));

  researchkey.push_back(key(L'\\', false));
  researchkey.push_back(key(L'n', true));

  init_curses();

  mousemask(ALL_MOUSE_EVENTS, NULL);

  curses_avail=true;
  cbreak();
  rootwin.keypad(true);

  global_bindings.set("Quit", quitkey);
  global_bindings.set("Cycle", key(L'\t', false));
  global_bindings.set("Refresh", KEY_CTRL(L'l'));

  global_bindings.set("Up", upkey);
  global_bindings.set("Down", downkey);
  global_bindings.set("LevelDown", key(L'J', false));
  global_bindings.set("LevelUp", key(L'K', false));
  global_bindings.set("Left", leftkey);
  global_bindings.set("Right", rightkey);
  global_bindings.set("HistoryNext", historynextkey);
  global_bindings.set("HistoryPrev", historyprevkey);
  global_bindings.set("Parent", key(L'^', false));
  global_bindings.set("PrevPage", ppagekey);
  global_bindings.set("NextPage", npagekey);
  global_bindings.set("Begin", homekey);
  global_bindings.set("End", endkey);
  global_bindings.set("Search", key(L'/', false));
  global_bindings.set("ReSearch", researchkey);
  global_bindings.set("DelBack", delbkey);
  global_bindings.set("DelForward", delfkey);

  global_bindings.set("DelEOL", KEY_CTRL(L'k'));
  global_bindings.set("DelBOL", KEY_CTRL(L'u'));

  global_bindings.set("Confirm", key(KEY_ENTER, true));
  global_bindings.set("Cancel", cancelkey);
  global_bindings.set("Undo", undokey);
  global_bindings.set("Help", helpkey);
  global_bindings.set("ToggleMenuActive", menutogglekey);
  global_bindings.set("PushButton", key(L' ', false));
  global_bindings.set("Yes", key(transcode(_("yes_key"))[0], false));
  global_bindings.set("No", key(transcode(_("no_key"))[0], false));

  global_bindings.set("ToggleExpanded", key(KEY_ENTER, true));
  global_bindings.set("ExpandAll", key(L'[', false));
  global_bindings.set("CollapseAll", key(L']', false));
  global_bindings.set("SelectParent", key(L'^', false));

  vs_editline::init_bindings();
  vs_menu::init_bindings();
  vs_menubar::init_bindings();
  vs_pager::init_bindings();
  vs_statuschoice::init_bindings();
  vs_table::init_bindings();
  vs_text_layout::init_bindings();
  vs_tree::init_bindings();

  set_style("Error",
	    style_fg(COLOR_WHITE)+style_bg(COLOR_RED)+style_attrs_on(A_BOLD));

  // The 'base' style for the display.
  set_style("Default",
	    style_fg(COLOR_WHITE)+style_bg(COLOR_BLACK));

  set_style("Header",
	    style_fg(COLOR_WHITE)+style_bg(COLOR_BLUE)+style_attrs_on(A_BOLD));
  set_style("Status",
	    style_fg(COLOR_WHITE)+style_bg(COLOR_BLUE)+style_attrs_on(A_BOLD));

  set_style("MenuEntry", style_fg(COLOR_WHITE)+style_bg(COLOR_BLUE));
  set_style("MenuBorder", style_fg(COLOR_WHITE)+style_bg(COLOR_BLUE)+style_attrs_on(A_BOLD));
  set_style("HighlightedMenuEntry", style_bg(COLOR_BLUE)+style_fg(COLOR_WHITE)+style_attrs_on(A_BOLD|A_REVERSE));
  set_style("DisabledMenuEntry",
	    style_fg(COLOR_BLACK)+style_bg(COLOR_BLUE)+style_attrs_on(A_DIM));
  set_style("MenuBar", style_fg(COLOR_WHITE)+style_bg(COLOR_BLUE)+style_attrs_on(A_BOLD));
  set_style("HighlightedMenuBar", style_fg(COLOR_BLUE)+style_bg(COLOR_WHITE));

  set_style("MultiplexTab", style_fg(COLOR_WHITE)+style_bg(COLOR_BLUE));
  set_style("MultiplexTabHighlighted", style_fg(COLOR_BLUE)+style_bg(COLOR_WHITE));

  // Edit lines will *always* appear white-on-black.
  set_style("EditLine", style_fg(COLOR_WHITE)+style_bg(COLOR_BLACK)+style_attrs_off(A_REVERSE));

  set_style("WidgetFrame", style_fg(COLOR_WHITE)+style_bg(COLOR_BLACK));

  set_style("TreeBackground", style());

  if(toplevel)
    vscreen_settoplevel(toplevel);

  pthread_mutexattr_t vscreen_mutex_attr;
  pthread_mutexattr_init(&vscreen_mutex_attr);
  pthread_mutexattr_settype(&vscreen_mutex_attr, PTHREAD_MUTEX_RECURSIVE);

  pthread_mutex_init(&vscreen_mutex, &vscreen_mutex_attr);

  pthread_mutexattr_destroy(&vscreen_mutex_attr);

  vscreen_install_sighandlers();
}

void vscreen_install_sighandlers()
{
  signal(SIGTERM, sigkilled);
  signal(SIGINT, sigkilled);
  signal(SIGSEGV, sigkilled);
  signal(SIGQUIT, sigkilled);
  signal(SIGABRT, sigkilled);
}

void vscreen_handleresize()
{
  toplevel->set_owner_window(NULL, 0, 0, 0, 0);
  resize();
  toplevel->set_owner_window(rootwin, 0, 0, rootwin.getmaxx(), rootwin.getmaxy());
  vscreen_redraw();
}

void vscreen_checktimeouts()
{
  list<timeout_info>::iterator i,j;
  for(i=timeouts.begin(); i!=timeouts.end(); i=j)
    {
      j=i;
      j++;
      timeval result,curtime;
      gettimeofday(&curtime, 0);

      if(timeval_subtract(&result, &i->activate_time, &curtime)==1 ||
	 result.tv_sec==0 && result.tv_usec<=10)
	// Fudge things a little.. (good idea or not?)
	{
	  i->f();
	  timeouts.erase(i);
	}
    }

  vscreen_tryupdate();
}

int vscreen_mindelay()
  // Returns the number of milliseconds to delay for the first timeout. (-1 for
  // infinite delay)
  // If a timeout should be done now, returns 0.
{
  bool found_one=false;
  timeval mintime;
  mintime.tv_sec=INT_MAX/1000;
  mintime.tv_usec=(INT_MAX%1000)*1000;

  timeval curtime;
  gettimeofday(&curtime, 0);

  list<timeout_info>::iterator i,j;
  for(i=timeouts.begin(); i!=timeouts.end(); i=j)
    {
      j=i;
      j++;
      timeval diff;
      if(timeval_subtract(&diff, &i->activate_time, &curtime)==1 ||
	 diff.tv_sec==0 && diff.tv_usec<=10)
	return 0;
      else
	{
	  if(diff.tv_sec<mintime.tv_sec ||
	     (diff.tv_sec==mintime.tv_sec && diff.tv_usec<mintime.tv_usec))
	  {
	    found_one=true;
	    mintime.tv_sec=diff.tv_sec;
	    mintime.tv_usec=diff.tv_usec;
	  }
	}
    }
  if(found_one)
    return mintime.tv_sec*1000+mintime.tv_usec/1000;
  else
    return -1;
}

void vscreen_updatecursor()
{
  needs_cursorupdate=true;
}

void vscreen_updatecursornow()
{
  if(toplevel->get_cursorvisible())
    {
      point p=toplevel->get_cursorloc();
      toplevel->win.move(p.y, p.x);
      toplevel->win.leaveok(false);
    }
  else
    toplevel->win.leaveok(true);

  needs_cursorupdate=false;
}

void vscreen_update()
{
  needs_update=true;
  needs_cursorupdate=true;
}

void vscreen_updatenow()
{
  if(toplevel)
    {
      toplevel->display(get_style("Default"));
      toplevel->sync();

      vscreen_updatecursornow();

      doupdate();
    }
  needs_update=false;
}

void vscreen_layoutnow()
{
  toplevel->do_layout();
  needs_layout=false;
}

void vscreen_tryupdate()
{
  if(needs_layout)
    vscreen_layoutnow();

  if(needs_update)
    vscreen_updatenow();

  if(needs_cursorupdate)
    {
      vscreen_updatecursornow();
      refresh();
    }
}

bool vscreen_poll()
  // FIXME: should I take the global lock?  Need to think about this before
  //       using libvscreen for threaded programming
{
  bool rval=false;

  sigset_t signals,prevsignals;

  sigemptyset(&signals);
  sigaddset(&signals, SIGWINCH);

  if(toplevel)
    {
      wint_t wch = 0;
      int status;

      do
	{
	  toplevel->win.nodelay(true);
	  status = toplevel->win.get_wch(&wch);
	  toplevel->win.nodelay(false);
	} while(wch==(wint_t) KEY_RESIZE);

      key k(wch, status == KEY_CODE_YES);

      sigprocmask(SIG_BLOCK, &signals, &prevsignals);

      if(status != ERR)
	{
	  rval = true;

	  if(wch == KEY_MOUSE)
	    {
	      MEVENT ev;
	      getmouse(&ev);

	      if(toplevel)
		{
		  toplevel->dispatch_mouse(ev.id, ev.x, ev.y, ev.z, ev.bstate);
		  main_hook();
		  vscreen_tryupdate();
		}
	    }
	  else if(global_bindings.key_matches(k, "Refresh"))
	    vscreen_redraw();
	  else if(toplevel)
	    {
	      toplevel->dispatch_key(k);
	      main_hook();
	      vscreen_tryupdate();
	    }
	}
    }

  vscreen_checktimeouts();

  sigprocmask(SIG_SETMASK, &signals, &prevsignals);

  return rval;
}

void vscreen_mainloop(int synch)
  // FIXME: can I restructure things so that this just calls vscreen_poll
  // repeatedly?  It'd be a lot cleaner..
{
  static int main_level=0;
  // HACK -- basically, delayed deletion avoids crashes by guaranteeing that
  // things only get deleted in the main loop.  But this breaks if sub-main-loops
  // delete them.  Solution: don't delete things in sub-main-loops..

  main_level++;

  sigset_t signals;

  sigemptyset(&signals);
  sigaddset(&signals, SIGWINCH);

  signal(SIGWINCH, sigresize);

  vscreen_acquirelock();

  // Do this first, so any pending updates from initialization routines get
  // flushed.
  vscreen_tryupdate();

  while(!should_exit && toplevel)
    {
      sigset_t prevsignals;

      assert(toplevel->get_win());

      int prevtimeout=toplevel->timeout(vscreen_mindelay());
      vscreen_widget *curwidget=toplevel;
      // Note that something bad might happen while we don't have the mutex..
      // OTOH, it's perfectly okay to dispatch the character to an unexpected
      // location; I just want to be sure that I reset the timeout of the
      // correct widget. (??)

      wint_t wch;
      int result;

      do
	{
	  vscreen_releaselock();
	  result=curwidget->get_win().get_wch(&wch);
	  vscreen_acquirelock();
	} while(wch==KEY_RESIZE);

      key k = key(wch, result == KEY_CODE_YES);

      // used when asynchronous resizing was dangerous
      //if(resized)
      //	{
      //  resized=false;
      //  vscreen_handleresize();
      //}

      curwidget->timeout(prevtimeout);

      sigprocmask(SIG_BLOCK, &signals, &prevsignals);
      // Oddly enough, the part I want to *protect* from signals is my own code
      // -- the ncurses code above is actually graceful about this situation!

      vscreen_checktimeouts();

      if(wch==KEY_MOUSE)
	{
	  MEVENT ev;
	  getmouse(&ev);

	  if(toplevel)
	    {
	      toplevel->dispatch_mouse(ev.id, ev.x, ev.y, ev.z, ev.bstate);
	      main_hook();
	      vscreen_tryupdate();
	    }
	}
      else if(global_bindings.key_matches(k, "Refresh"))
	vscreen_redraw();
      else if(result!=ERR && toplevel)
	{
	  toplevel->dispatch_key(k);
	  main_hook();
	  vscreen_tryupdate();
	}

      if(main_level==1)
	vscreen_widget::handle_pending_deletes();

      sigprocmask(SIG_SETMASK, &prevsignals, NULL);
    }

  should_exit=false;

  vscreen_releaselock();

  main_level--;
}

void vscreen_exitmain()
{
  should_exit=1;
}

void vscreen_suspend()
{
  if(toplevel)
    toplevel->set_owner_window(NULL, 0, 0, 0, 0);
  endwin();
  curses_avail=false;
}

void vscreen_resume()
{
  curses_avail=true;
  if(toplevel)
    {
      toplevel->set_owner_window(rootwin, 0, 0, rootwin.getmaxx(), rootwin.getmaxy());
      toplevel->display(get_style("Default"));
      toplevel->sync();
      doupdate();
    }
  else
    refresh();
}

void vscreen_queuelayout()
{
  needs_layout=true;
  needs_update=true;
  needs_cursorupdate=true;
}

void vscreen_redraw()
{
  if(toplevel)
    {
      toplevel->focussed();
      toplevel->get_win().touch();
      toplevel->get_win().clearok(true);
      toplevel->do_layout();
      toplevel->display(get_style("Default"));
      vscreen_updatecursornow();
      toplevel->sync();
      doupdate();
      toplevel->get_win().clearok(false);
    }
}

int vscreen_addtimeout(const sigc::slot0<void> &slot, int msecs)
{
  if(msecs<0)
    return -1;

  int rval=ntimeouts++;
  timeval activate_time;
  gettimeofday(&activate_time, 0);
  activate_time.tv_sec+=msecs/1000;
  activate_time.tv_usec+=(msecs%1000)*1000;
  while(activate_time.tv_usec>1000*1000)
    // Should only run through once
    {
      activate_time.tv_sec++;
      activate_time.tv_usec-=1000*1000;
    }

  timeouts.push_back(timeout_info(slot, rval, activate_time));
  return rval;
}

void vscreen_deltimeout(int num)
{
  for(list<timeout_info>::iterator i=timeouts.begin(); i!=timeouts.end(); i++)
    {
      if(i->num==num)
	{
	  timeouts.erase(i);
	  return;
	}
    }
}

void vscreen_acquirelock()
{
#ifdef HAVE_LIBPTHREAD
  pthread_mutex_lock(&vscreen_mutex);
#endif
}

void vscreen_releaselock()
{
#ifdef HAVE_LIBPTHREAD
  pthread_mutex_unlock(&vscreen_mutex);
#endif
}

